// File: ClassifierTest.cpp
// Purpose: Implementation of the ClassifierTest class and its factory class
// Remarks: This class is an abstract base class.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 11/15/2007 3:11:46 PM

#include "ClassifierTest.h"

using namespace std;

namespace idl
{
	// Clear the records of the tests on all available parameters
	void ClassifierTest::reset()
	{
		testCount = 0;
		errorSum.zeros();
		errorSumOfSquares.zeros();
	}

	// Clear the records of the tests on all available parameters
	void ClassifierTest::reset(unsigned int numParameterSets)
	{
		testCount = 0;

		if (0 == numParameterSets)
		{
			errorSum.clear();
			errorSumOfSquares.clear();
		}
		else
		{
			errorSum.zeros(numParameterSets, 1);
			errorSumOfSquares.zeros(numParameterSets, 1);
		}
	}

	// Get the cross-validation error of a parameter set
	double ClassifierTest::getError(unsigned int idx) const
	{
		return errorSum(idx) / static_cast<double>(testCount);
	}

	// Get the standard deviation of the cross-validation error
	double ClassifierTest::getErrorStdDev(unsigned int idx) const
	{
		if (1 == testCount)
		{
			return 0.0;
		}
		else
		{
			double error = getError(idx);
			double v = errorSumOfSquares(idx) / static_cast<double>(testCount);
			v -= error * error;
			return (v > 0.0) ? sqrt(v) : 0.0;
		}
	}

	// Get the information about the minimum cross-validation error
	void ClassifierTest::getMinErrorInfo(unsigned int &idx, double &error, double &stddev) const
	{
		error = errorSum.min();

		// Get the index of the parameter set achieving minimum error
		idx = 0;
		while (errorSum(idx) != error)
		{
			++idx;
		}

		// Get the minimum cross-validation error
		error /= static_cast<double>(testCount);

		// Get the standard deviation of the corresponding cross-validation error
		stddev = getErrorStdDev(idx);
	}

	// Get the map for the ClassifierTest creators
	ClassifierTestFactory::MapType & ClassifierTestFactory::getCreatorMap()
	{
		// Pointer to the the ClassifierTest creator map
		static MapType *creatorMap = new MapType;

		return *creatorMap;
	}

	// Register a ClassifierTest creator
	bool ClassifierTestFactory::registerCreator(const string &id,
		ClassifierTestFactory::CreatorType creator)
	{
		MapType &creatorMap = getCreatorMap();

		if (creatorMap.find(id) == creatorMap.end())
		{
			creatorMap[id] = creator;
			return true;
		}
		else
		{
			return false;
		}
	}

	// Create an instance of the specified ClassifierTest subclass
	ClassifierTest *ClassifierTestFactory::create(const string &id)
	{
		MapType &creatorMap = getCreatorMap();

		MapType::const_iterator itr = creatorMap.find(id);

		return (itr == creatorMap.end()) ? NULL : (itr->second)();
	}
}
