// File: ConfigParser.cpp
// Purpose: Implementation of the ConfigParser class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 9/14/2007 12:12:03 PM

#include <cstring>
#include <string>
#include <fstream>
#include <iostream>
#include "ConfigParser.h"

using namespace std;

namespace idl
{
	// Destructor
	ConfigParser::~ConfigParser()
	{
		// Delete the maps storing classifier configurations
		for (vector<ConfigMap *>::iterator itr = classifierConfig.begin();
			 itr != classifierConfig.end(); ++itr)
		{
			delete *itr;
		}
	}

	// Clear the current configurations
	void ConfigParser::clear()
	{
		// Clear the map storing test configurations
		testConfig.clear();

		// Delete the maps storing classifier configurations
		for (vector<ConfigMap *>::iterator itr = classifierConfig.begin();
			 itr != classifierConfig.end(); ++itr)
		{
			delete *itr;
		}

		classifierConfig.clear();
	}

	// Get the number of the classifiers to test
	unsigned int ConfigParser::getClassifierNum() const
	{
		return static_cast<unsigned int>(classifierConfig.size());
	}

	// Print the command line usage
	void ConfigParser::printUsage(const char *progName) const
	{
		cout << "Usage: " << progName
			 << " -c file\tspecify the configuration file" << endl;
	}

	// Parse the command line
	bool ConfigParser::parseCommandLine(int argc, char *argv[])
	{
		if (argc < 3)
		{
			cerr << "Error: Incorrect number of arguments" << endl;
			printUsage(argv[0]);
			return false;
		}

		clear();

		int pos = 1;

		while (pos < argc)
		{
			if (0 == strcmp(argv[pos], "-c"))
			{
				if (++pos >= argc)
				{
					cerr << "Error: Missing arguments" << endl;
					printUsage(argv[0]);
					return false;
				}

				if (!parseConfigFile(argv[pos++])) return false;
			}
			else
			{
				cerr << "Error: Unknown option" << endl;
				printUsage(argv[0]);
				return false;
			}
		}

		return true;
	}

	// Parse the config file
	bool ConfigParser::parseConfigFile(const char *file)
	{
		ifstream configFile(file);

		if (!configFile)
		{
			cerr << "Error: Cannot open file " << file << endl;
			return false;
		}

		clear();

		string line;
		string key;
		string value;

		// Parse test configuration
		while (!configFile.eof())
		{
			getline(configFile, line);
			removeComments(line);
			splitKeyValue(line, key, value);

			if ("Classifier" == key) break;

			if (key.empty() || value.empty()) continue;

			testConfig[key] = value;
		}

		// Parse parameters for each classifier
		if ("Classifier" != key || value.empty())
		{
			cerr << "Error: Missing configuration parameters" << endl;
			return false;
		}

		ConfigMap *ptClassifierConfig = new ConfigMap;
		(*ptClassifierConfig)[key] = value;
		classifierConfig.push_back(ptClassifierConfig);

		while (!configFile.eof())
		{
			getline(configFile, line);
			removeComments(line);
			splitKeyValue(line, key, value);

			if ("Classifier" == key)
			{
				if (value.empty())
				{
					cerr << "Error: Missing configuration parameters" << endl;
					return false;
				}

				ptClassifierConfig = new ConfigMap;
				(*ptClassifierConfig)[key] = value;
				classifierConfig.push_back(ptClassifierConfig);
			}
			else
			{
				if (key.empty() || value.empty()) continue;

				(*ptClassifierConfig)[key] = value;
			}
		}

		return true;
	}

	// Remove comments and trailing space characters
	void ConfigParser::removeComments(string &str)
	{
		// Find comments
		string::size_type pos = str.find('#');

		// Remove comments
		if (string::npos != pos)
		{
			str.erase(str.begin() + pos, str.end());
		}

		// Find trailing space characters
		pos = str.find_last_not_of(" \t");

		// Remove trailing space characters
		if (string::npos == pos)
		{
			str.clear();
		}
		else
		{
			str.erase(str.begin() + pos + 1, str.end());
		}
	}

	// Split a line into key and value
	void ConfigParser::splitKeyValue(const string &line, string &key, string &value)
	{
		// Find the first space character
		string::size_type pos = line.find_first_of(" \t");

		if (string::npos == pos)
		{
			key = line;
			value.clear();
		}
		else
		{
			if (0 == pos)
			{
				key.clear();
			}
			else
			{
				key.assign(line, 0, pos);
			}

			pos = line.find_first_not_of(" \t", pos);

			if (string::npos == pos)
			{
				value.clear();
			}
			else
			{
				value.assign(line.begin() + pos, line.end());
			}
		}
	}

	// Split a line into fields separated by a specific delimiter
	void ConfigParser::splitFields(const string &line, vector<string> &fields, char delimiter)
	{
		fields.clear();

		for (string::size_type pos1 = 0, pos2; pos1 < line.size(); pos1 = pos2 + 1)
		{
			pos2 = line.find(delimiter, pos1);

			if (string::npos == pos2)
			{
				fields.push_back(line.substr(pos1));
				break;
			}
			else if (pos2 > pos1)
			{
				fields.push_back(line.substr(pos1, pos2 - pos1));
			}
		}
	}

	// Get a string value from a config map
	bool ConfigParser::getString(const ConfigMap &map, const string &key, string &value)
	{
		ConfigMap::const_iterator itr = map.find(key);

		if (itr == map.end()) return false;

		value = itr->second;

		return true;
	}

	// Get a string value
	bool ConfigParser::getString(int secIdx, const string &key, string &value) const
	{
		return (secIdx < 0) ? getString(testConfig, key, value)
			: getString((*(classifierConfig[secIdx])), key, value);
	}

	// Get a string value
	bool ConfigParser::getString(int secIdx, const char *key, string &value) const
	{
		return getString(secIdx, string(key), value);
	}

	// Get an integer value
	bool ConfigParser::getInt(int secIdx, const string &key, int &value) const
	{
		return getT<int>(secIdx, key, value);
	}

	// Get an integer value
	bool ConfigParser::getInt(int secIdx, const char *key, int &value) const
	{
		return getInt(secIdx, string(key), value);
	}

	// Get an unsigned integer value
	bool ConfigParser::getUInt(int secIdx, const string &key, unsigned int &value) const
	{
		return getT<unsigned int>(secIdx, key, value);
	}

	// Get an unsigned integer value
	bool ConfigParser::getUInt(int secIdx, const char *key, unsigned int &value) const
	{
		return getUInt(secIdx, string(key), value);
	}

	// Get a real value
	bool ConfigParser::getReal(int secIdx, const string &key, double &value) const
	{
		return getT<double>(secIdx, key, value);
	}

	// Get a real value
	bool ConfigParser::getReal(int secIdx, const char *key, double &value) const
	{
		return getReal(secIdx, string(key), value);
	}

	// Get an integer vector
	bool ConfigParser::getIntVec(int secIdx, const string &key, vector<int> &vec) const
	{
		return getTVec<int>(secIdx, key, vec);
	}

	// Get an integer vector
	bool ConfigParser::getIntVec(int secIdx, const char *key, vector<int> &vec) const
	{
		return getIntVec(secIdx, string(key), vec);
	}

	// Get an unsigned integer vector
	bool ConfigParser::getUIntVec(int secIdx, const string &key, vector<unsigned int> &vec) const
	{
		return getTVec<unsigned int>(secIdx, key, vec);
	}

	// Get an unsigned integer vector
	bool ConfigParser::getUIntVec(int secIdx, const char *key, vector<unsigned int> &vec) const
	{
		return getUIntVec(secIdx, string(key), vec);
	}

	// Get a real vector
	bool ConfigParser::getRealVec(int secIdx, const string &key, vector<double> &vec) const
	{
		return getTVec<double>(secIdx, key, vec);
	}

	// Get a real vector
	bool ConfigParser::getRealVec(int secIdx, const char *key, vector<double> &vec) const
	{
		return getRealVec(secIdx, string(key), vec);
	}
}
