// File: LimeSolver.h
// Purpose: Definition of the LimeSolver class
// Remarks: LimeSolver will call MOSEK (http://www.mosek.com/) to solve
//			the following convex optimization problem:
//
//				min  1/2 * w' * Q * w + q' * w - lambda * H(w)
//				s.t. w >= 0, 1' * w = 1
//
//			The standard form of a linearly constrained nonlinear optimization
//			problem specified by MOSEK is:
//
//				min  f(x) + c' * x
//				s.t. l_c <= A * x <= u_c
//					 l_x <= x <= u_x
//
//			Here we have:
//
//				f(x) = 1/2 * x' * Q * x - lambda * H(x)
//				c = q, A = 1', l_c = u_c = 1, l_x = 0, u_x = inf
//
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 7/23/2007 10:22AM

#ifndef IDL_LIMESOLVER_H
#define IDL_LIMESOLVER_H

#include "mosek.h" // MOSEK header file

namespace idl
{
	/**
	 * Class for computing the LIME weights.
	 * @remarks		LimeSolver will call MOSEK to solve the LIME problem (a convex
	 *				optimization problem).
	 * @attention	LimeSolver is a singleton implemented by static members due to
	 *				the fact that the LimeSolver class provides several callback
	 *				functions for MOSEK.
	 * @warning		No instance of the LimeSolver class is allowed to be created. 
	 * @see			<a href="http://www.mosek.com/">MOSEK</a>
	 */
	class LimeSolver
	{
	public:
		/**
		 * Solve the convex optimization problem for LIME.
		 * @param[in]	numNeighbors	number of neighbors
		 * @param[in]	QMtx			pointer to matrix Q
		 * @param[in]	qVec			pointer to vector q
		 * @param[in]	lambdaVal		lambda value
		 * @param[out]	weights			pointer to the buffer to store the results
		 * @return		true if success; false if failure
		 */
		static bool solve(int		numNeighbors,
						  double	*QMtx,
						  double	*qVec,
						  double	lambdaVal,
						  double	*weights);

		/**
		 * Print the messages from MOSEK on the screen.
		 * @warning		DO NOT call this function. As a callback function
		 *				for MOSEK, this function is supposed to be called
		 *				by MOSEK ONLY.
		 */
		static void MSKAPI mosekPrintLog(void *handle, char str[]);

		/**
		 * Inform MOSEK of the problem structure and sparsity.
		 * @warning		DO NOT call this function. As a callback function
		 *				for MOSEK, this function is supposed to be called
		 *				by MOSEK ONLY.
		 */
		static int MSKAPI mosekNlSpar(void	*nlhandle,
									  int	*numgrdobjnz,
									  int	*grdobjsub,
									  int	i,
									  int	*convali,
									  int	*grdconinz,
									  int	*grdconisub,
									  int	yo,
									  int	numycnz,
									  int	ycsub[],
									  int	maxnumhesnz,
									  int	*numhesnz,
									  int	*hessubi,
									  int	*hessubj);

		/**
		 * Evaluate the nonlinear function (including its gradient and Hessian)
		 * and return the requested information to MOSEK.
		 * @warning		DO NOT call this function. As a callback function
		 *				for MOSEK, this function is supposed to be called
		 *				by MOSEK ONLY.
		 */
		static int MSKAPI mosekNlEval(void		*nlhandle,
									  double	*xx,
									  double	yo,
									  double	*yc,
									  double	*objval,
									  int		*numgrdobjnz,
									  int		*grdobjsub,
									  double	*grdobjval,
									  int		numi,
									  int		*subi,
									  double	*conval,
									  int		*grdconptrb,
									  int		*grdconptre,
									  int		*grdconsub,
									  double	*grdconval,
									  double	*grdlag,
									  int		maxnumhesnz,
									  int		*numhesnz,
									  int		*hessubi,
									  int		*hessubj,
									  double	*hesval);

	private:
		/**
		 * Constructor.
		 * @remarks		To make the constructor private is to prevent
		 *				the creation of any LimeSolver object.
		 */
		LimeSolver() {};

		/**
		 * Set the information about the optimization problem.
		 * @param[in]	numNeighbors	number of neighbors
		 * @return		true if success; false if failure
		 */
		static bool setProbInfo(int numNeighbors);

		/** Clear the information about the optimization problem. */
		static void clearProbInfo();

		/**
		 * Set the parameters of the optimization problem.
		 * @param[in]	QMtx		pointer to matrix Q
		 * @param[in]	qVec		pointer to vector q
		 * @param[in]	lambdaVal	lambda value
		 * @return		true if success; false if failure
		 */
		static bool setProbPara(double *QMtx, double *qVec, double lambdaVal);

		/** Clear the parameters of the optimization problem. */
		static void clearProbPara();

#ifdef _DEBUG
		/** Print the parameters of the optimization problem on the screen. */
		static void printProbPara();
#endif // _DEBUG

		/**
		 * Compute f(x), the nonlinear part of the objective function.
		 * @param[in]	x	pointer to vector x
		 * @return		value of f(x)
		 */
		static double f(double *x);

		/**
		 * Compute the gradient of f(x).
		 * @param[in]	x	pointer to vector x
		 * @param[in]	i	index of the component
		 * @return		value of the ith component of the gradient
		 */
		static double f_grad(double *x, int i);

		/**
		 * Compute the Hessian of f(x).
		 * @param[in]	x	pointer to vector x
		 * @param[in]	i	index of the row
		 * @param[in]	j	index of the column
		 * @return		value of the element in the ith row and jth column of
		 *				the Hessian matrix
		 */
		static double f_hessian(double *x, int i, int j);

	private:
		static const int NUM_EQ_CON; /**< Number of equality constraints. */

		static MSKboundkeye bkc[1]; /**< Bound keys for the constraints. */
		static double blc[1];		/**< Lower bounds for the constraints. */
		static double buc[1];		/**< Upper bounds for the constraints. */

		static int *ptrb;	/**< Start positions of the constraint values. */
		static int *ptre;	/**< End positions of the constraint values. */
		static int *sub;	/**< Constraint subscripts. */

		static MSKboundkeye *bkx; /**< Bound keys for the variables. */

		static double *blx;	/**< Lower bounds for the variables. */
		static double *bux;	/**< Upper bounds for the variables. */
		static double *val;	/**< Constraint values. */
		static double *sol;	/**< Buffer to store the solution. */

		static int numVariables; /**< Number of varialbes. */

		/**
		 * Number of non-zero elements in the lower triangle of
		 * the Hessian matrix of f(x).
		 */
		static int numHesNz;

		static double *Q;		/**< Pointer to matrix Q. */
		static double *q;		/**< Pointer to vector q. */
		static double lambda;	/**< Lambda value. */
	};
}

#endif // IDL_LIMESOLVER_H
