// File: LocalSDABNTest.cpp
// Purpose: Implementation of the LocalSDABNTest class
// Remarks: This class is designed as an adapter for the WeightedKNN class
//			to enable testing on multiple sets of parameters.
// Author: Ingrid Liu(ingrid@hkn.eecs.berkeley.edu)
// Last modified: 5/27/2008 4:24:41 PM

#include <cmath>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "LocalSDABNTest.h"

using namespace std;

namespace idl
{
	const string LocalSDABNTest::_ID = "lSDA-BN"; // Define the class ID

	// Register the class instantiator with the factory
	const bool LocalSDABNTest::_REGISTERED =
		ClassifierTestFactory::registerCreator(LocalSDABNTest::_ID, &LocalSDABNTest::create);

	// Destructor
	LocalSDABNTest::~LocalSDABNTest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Configure the test for a weighted k-NN classifier
	bool LocalSDABNTest::configure(ConfigParser &config, int secIdx)
	{
		
		// Get neighborhood sizes to cross-validate
		if (!config.getUIntVec(secIdx, "NeighborhoodSize", neighborhoodSizes))
		{
			cerr << "Error: Missing or incorrect neighborhood size for lSDA classifier "
				 << secIdx << endl;
			return false;
		}

		// Check the retrieved neighborhood sizes
		if (!checkNeighborhoodSizes())
		{
			cerr << "Error: Incorrect neighborhood size for lSDA classifier "
				 << secIdx << endl;
			return false;
		}

		reset(static_cast<unsigned int>(1));

		return true;
	}

	// Test the Local SDA-BN classifier on all available parameters
	bool LocalSDABNTest::test(const SimData &data, ofstream &logFile)
	{
		//std::cout << "Testing... \n";
		// Log the weight mode
		logFile << "Local SDA-BN classifier" << endl
				<< "Neighborhood size\tRegParam\tError rate" << endl;

		unsigned int i = 0; // parameter set count

		vector<int> results(data.getTestSetSize());

		//for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
		//	 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
		//{
			lsda.setNeighborhoodSizes(neighborhoodSizes);

			//for (vector<double>::const_iterator itrRegParam = regParams.begin();
			//	 itrRegParam != regParams.end(); ++itrRegParam)
			//{
				//std::cout << "Testing ... neighborhood size: " << *itrNeighborhoodSize << " reg: " << *itrRegParam << std::endl;
			//	lsda.setRegPara(*itrRegParam);
				if (!lsda.classify(data, results)) return false;

				double errorRate = data.getClassificationError(results);
				
				// Log the error rate
				logFile << (errorRate * 100.0) << "%" << endl;

				errorSum(i) += errorRate;
				errorSumOfSquares(i++) += (errorRate * errorRate);
			//}
		//}

		logFile << endl;

		++testCount;
		return true;
	}

	// Check the legitimacy of the neighborhood sizes
	bool LocalSDABNTest::checkNeighborhoodSizes()
	{
		sort(neighborhoodSizes.begin(), neighborhoodSizes.end());

		return neighborhoodSizes.front() > 0;
	}

	// Clear the results with the best parameters
	void LocalSDABNTest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void LocalSDABNTest::selectBestPara()
	{
		
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet

		if (0 == testCount)
		{
			// No cross-validation performed
			//latestResult.regParam = regParams.front();
			//latestResult.neighborhoodSize = neighborhoodSizes.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);

			// Get the neighborhood size corresponding to the minimum cross-validation error
			//latestResult.neighborhoodSize = neighborhoodSizes[i];

		}

		resultsBestPara.push_back(latestResult);
	}

	// Test the weighted k-NN classifier with the best parameters
	bool LocalSDABNTest::testBestPara(const SimData &data, ofstream &logFile)
	{
		// Log the weight mode
		logFile << "Local SDA-BN classifier" << endl;

		//unsigned int k = (resultsBestPara.back()).neighborhoodSize;
		lsda.setNeighborhoodSizes(neighborhoodSizes);

		vector<int> results(data.getTestSetSize());
		RealMatrix probMatrix;
		if (!lsda.classify(data, results, probMatrix)) return false;
		
		logFile << "Classification probabilities:" << endl;
		for (unsigned int i = 0; i < probMatrix.getRowNum(); ++i)
		{
			for (unsigned int j = 0; j < probMatrix.getColumnNum(); ++j)
			{
				logFile << probMatrix(i,j) << " ";
			}
			logFile << endl;
		}
		logFile << "End of classification probabilities." << endl;

		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;

		// Log the error rate
		logFile << "Error rate" << endl
				<< (errorRate * 100.0) << "%" << endl << endl;

		return true;
	}

	// Report the test results of all available parameters
	void LocalSDABNTest::reportResults(ofstream &file, bool verbose)
	{
		file << "Local SDA-BN classifier" << endl;

		
		// Report the best result only
		file << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
			 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl;
		

		file << endl;
	}

	// Report the test results with the best parameters
	void LocalSDABNTest::reportBestParaResults(ofstream &file, bool verbose)
	{
		file << "Local SDA-BN classifier" << endl;

		if (verbose)
		{
			file << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}
		else
		{
			file << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Summarize the results of all the tests with the best parameters
	void LocalSDABNTest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		file << "Local SDA-BN classifier" << endl;

		if (verbose)
		{
			file << "Error rate" << endl;
		}

		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		//RealMatrix kVec(1, n);
		//RealMatrix regParamVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			//kVec(i) = static_cast<double>(itr->neighborhoodSize);
			//regParamVec(i) = itr->regParam;
			errorVec(i) = itr->testError;

			if (verbose)
			{
				file << (itr->testError * 100.0) << "%" << endl;
			}
		}
		
		file << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
