// File: NearestMeanTest.cpp
// Purpose: Implementation of the NearestMeanTest class
// Remarks: This class is designed as an adapter for the NearestMean class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/27/2008 4:20:14 PM

#include <climits>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "NearestMeanTest.h"

using namespace std;

namespace idl
{
	const string NearestMeanTest::_ID = "nearest-mean"; // Define the class ID

	// Register the class instantiator with the factory
	const bool NearestMeanTest::_REGISTERED =
		ClassifierTestFactory::registerCreator(NearestMeanTest::_ID, &NearestMeanTest::create);

	// Destructor
	NearestMeanTest::~NearestMeanTest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Configure the test for a nearest-mean classifier
	bool NearestMeanTest::configure(ConfigParser &config, int secIdx)
	{
		string strMode;

		// Get the mode for the nearest-mean classifier
		if (!config.getString(secIdx, "Mode", strMode))
		{
			cerr << "Missing mode for nearest-mean classifier " << secIdx << endl;
			return false;
		}
		if ("global" == strMode)
		{
			nm.setGlobalFlag(true);
			neighborhoodSizes.clear();
			reset(1);
		}
		else if ("local" == strMode)
		{
			nm.setGlobalFlag(false);

			// Get neighborhood sizes to cross-validate
			if (!config.getUIntVec(secIdx, "NeighborhoodSize", neighborhoodSizes))
			{
				cerr << "Error: Missing or incorrect neighborhood size for nearest-mean classifier "
					 << secIdx << endl;
				return false;
			}

			// Check the retrieved neighborhood sizes
			if (!checkNeighborhoodSizes())
			{
				cerr << "Error: Incorrect neighborhood size for nearest-mean classifier "
					 << secIdx << endl;
				return false;
			}

			reset(static_cast<unsigned int>(neighborhoodSizes.size()));
		}
		else
		{
			cerr << "Unknown mode for nearest-mean classifier " << secIdx << endl;
			return false;
		}

		return true;
	}

	// Test the nearest-mean classifier on all available parameters
	bool NearestMeanTest::test(const SimData &data, ofstream &logFile)
	{
		vector<int> results(data.getTestSetSize());

		// Log the nearest-mean classifier mode
		logFile << "Nearest-mean classifier" << endl << "Mode: ";

		if (neighborhoodSizes.empty())
		{
			logFile << "global" << endl << "Error rate: ";
			if (!nm.classify(data, results)) return false;

			double errorRate = data.getClassificationError(results);

			// Log the error rate
			logFile << (errorRate * 100.0) << "%" << endl;

			errorSum(0) += errorRate;
			errorSumOfSquares(0) += (errorRate * errorRate);
		}
		else
		{
			logFile << "local" << endl << "Neighborhood size\tError rate" << endl;

			unsigned int i = 0;
			for (vector<unsigned int>::const_iterator itr = neighborhoodSizes.begin();
				 itr != neighborhoodSizes.end(); ++itr)
			{
				nm.setNeighborhoodSize(*itr);
				if (!nm.classify(data, results)) return false;

				double errorRate = data.getClassificationError(results);

				// Log the error rate
				logFile << *itr << "\t" << (errorRate * 100.0) << "%" << endl;

				errorSum(i) += errorRate;
				errorSumOfSquares(i++) += (errorRate * errorRate);
			}
		}

		logFile << endl;

		++testCount;

		return true;
	}

	// Check the legitimacy of the neighborhood sizes
	bool NearestMeanTest::checkNeighborhoodSizes()
	{
		sort(neighborhoodSizes.begin(), neighborhoodSizes.end());

		return neighborhoodSizes.front() > 0;
	}

	// Clear the results with the best parameters
	void NearestMeanTest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void NearestMeanTest::selectBestPara()
	{
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet

		if (0 == testCount)
		{
			// No cross-validation performed
			latestResult.neighborhoodSize =
				neighborhoodSizes.empty() ? UINT_MAX : neighborhoodSizes.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);

			// Get the neighborhood size corresponding to the minimum cross-validation error
			latestResult.neighborhoodSize =
				neighborhoodSizes.empty() ? UINT_MAX : neighborhoodSizes[i];
		}

		resultsBestPara.push_back(latestResult);
	}

	// Test the nearest-mean classifier with the best parameters
	bool NearestMeanTest::testBestPara(const SimData &data, ofstream &logFile)
	{
		// Log the nearest-mean classifier mode
		logFile << "Nearest-mean classifier" << endl << "Mode: ";

		unsigned int k = (resultsBestPara.back()).neighborhoodSize;

		if (UINT_MAX == k)
		{
			logFile << "global" << endl;
			nm.setGlobalFlag(true);
		}
		else
		{
			logFile << "local" << endl << "Neighborhood size: " << k << endl;
			nm.setGlobalFlag(false);
			nm.setNeighborhoodSize(k);
		}

		vector<int> results(data.getTestSetSize());
		if (!nm.classify(data, results)) return false;

		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;

		// Log the error rate
		logFile << "Error rate: " << (errorRate * 100.0) << "%" << endl << endl;

		return true;
	}

	// Report the test results of all available parameters
	void NearestMeanTest::reportResults(ofstream &file, bool verbose)
	{
		file << "Nearest-mean classifier" << endl << "Mode: ";

		if (neighborhoodSizes.empty())
		{
			file << "global" << endl
				 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl;
		}
		else
		{
			file << "local" << endl;

			if (verbose)
			{
				// Report the results of all the parameter sets
				file << "Neighborhood size\tError rate\tStandard deviation" << endl;

				unsigned int i = 0;
				for (vector<unsigned int>::const_iterator itr = neighborhoodSizes.begin();
					 itr != neighborhoodSizes.end(); ++itr)
				{
					file << *itr << "\t" << (getError(i) * 100.0) << "%\t" << (getErrorStdDev(i) * 100.0) << "%" << endl;
					++i;
				}
			}
			else
			{
				// Report the best result only
				file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
					 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
					 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl;
			}
		}

		file << endl;
	}

	// Report the test results with the best parameters
	void NearestMeanTest::reportBestParaResults(ofstream &file, bool verbose)
	{
		file << "Nearest-mean classifier" << endl << "Mode: ";

		if (neighborhoodSizes.empty())
		{
			file << "global" << endl;
		}
		else
		{
			file << "local" << endl
				 << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl;
		}

		if (verbose)
		{
			file << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}
		else
		{
			file << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}

		file << endl;
	}
	
	// Summarize the results of all the tests with the best parameters
	void NearestMeanTest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		file << "Nearest-mean classifier" << endl << "Mode: ";

		if (neighborhoodSizes.empty())
		{
			file << "global" << endl;
		}
		else
		{
			file << "local" << endl;

			if (verbose)
			{
				file << "Neighborhood size\tError rate" << endl;
			}
		}

		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		RealMatrix kVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			kVec(i) = static_cast<double>(itr->neighborhoodSize);
			errorVec(i) = itr->testError;

			if (!neighborhoodSizes.empty() && verbose)
			{
				file << itr->neighborhoodSize << "\t" << (itr->testError * 100.0) << "%" << endl;
			}
		}

		if (!neighborhoodSizes.empty())
		{
			file << "Average neighborhood size: " << kVec.mean()
				 << " (" << kVec.std() << ")" << endl;
		}

		file << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
