// File: PSVM.h
// Purpose: Definition of the PSVM class
// Remarks: This class is designed for PSVM classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 6/10/2008 6:42:32 PM

#ifndef IDL_PSVM_H
#define IDL_PSVM_H

#include <cstring>
#include <string>
#include <vector>
#include "common.h"
#include "RealMatrix.h"
#include "Classifier.h"
#include "SimData.h"

namespace idl
{
	/**
	 * Class for the PSVM classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class PSVM : public Classifier
	{
	public:
		/** Structure type for PSVM parameters. */
		struct ParaType
		{
			std::string	psvmBinary;			/**< Path of the PSVM binary. */
			std::string workingDirectory;	/**< Path of the working directory. */
			double C;						/**< Hyperparameter for PSVM. */
			double epsilon;					/**< Hyperparameter for PSVM. */
		};

		/** Default constructor. */
		PSVM()
		{
#ifdef LINUX
			parameters.psvmBinary = "psvm";
			parameters.workingDirectory = "/tmp/psvm/";
#else // WINDOWS
			parameters.psvmBinary = "psvm.exe";
			parameters.workingDirectory = "psvm\\";
#endif // LINUX
			parameters.C = 1000;
			parameters.epsilon = 0.01;
		}

		/**
		 * Constructor.
		 * @param	para	parameters for the PSVM classifier
		 */
		PSVM(const ParaType &para)
		{
			parameters = para;
		}

		~PSVM() {}	/**< Destructor. */

		/**
		 * Set the path of the PSVM binary.
		 * @param[in]	binaryPath		path of the PSVM binary
		 */
		void setBinaryPath(const std::string &binaryPath)
		{
			parameters.psvmBinary = binaryPath;
		}

		/**
		 * Set the path of the working directory.
		 * @param[in]	workingDirectory	path of the working directory
		 */
		void setWorkingDirectory(const std::string &workingDirectory)
		{
			parameters.workingDirectory = workingDirectory;
		}

		/**
		 * Set one of the floating-point parameters.
		 * @param[in]	name	name of the floating-point parameter
		 * @param[in]	value	value of the floating-point parameter
		 */
		void setPara(const char *name, double value)
		{
			if (0 == ::strcmp(name, "C"))
			{
				parameters.C = value;
			}
			else if (0 == ::strcmp(name, "epsilon"))
			{
				parameters.epsilon = value;
			}
		}

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		similarity data set
		 * @param[out]	results		vector to store the estimated class labels
		 * @return		true if success; false if failure
		 */
		virtual bool classify(const SimData &data, std::vector<int> &results);

	private:
		/** Structure for binary classification model. */
		struct BinaryModel
		{
			unsigned int classIdx1;	/**< Index of the 1st class (label: "+1"). */
			unsigned int classIdx2;	/**< Index of the 2nd class (label: "-1"). */
			std::string modelName;	/**< Path of the model file generated by the PSVM binary. */
		};

		/**
		 * Train the PSVM.
		 * @param[in]	trainingSamples		similarity matrix of the training samples
		 * @param[in]	labels				labels of the training samples
		 * @return		true if success; false if failure
		 */
		bool train(const RealMatrix &trainingSimMat, const std::vector<int> &labels);

		/*
		 * Predict the label of a test sample.
		 * @param[in]	testSimVec			test similarity vector
		 * @param[out]	predictedLabel		label predicted for the test sample
		 * @return		true if success; false if failure
		 */
		bool predict(const RealMatrix &testSimVec, int &predictedLabel);

		/**
		 * Predict the labels of the test samples.
		 * @param[in]	testSims			test similarity vectors (row vectors)
		 * @param[out]	predictedLabels		labels predicted for the test samples
		 * @return		true if success; false if failure
		 */
		bool predict(const RealMatrix &testSims, std::vector<int> &predictedLabels);

		/**
		 * Write the training data to a data file.
		 * @param[in]	simMat		similarity matrix between the training samples
		 * @param[in]	modelInfo	information of the binary classification model
		 * @return		true if success; false if failure
		 */
		bool writeTrainingDataFile(const RealMatrix &simMat, const BinaryModel &modelInfo);

		/**
		 * Write the test data to a data file.
		 * @param[in]	testSims	test similarity vectors (row vectors)
		 * @param[in]	modelInfo	information of the binary classification model
		 * @return		true if success; false if failure
		 */
		bool writeTestDataFile(const RealMatrix &testSims, const BinaryModel &modelInfo);

		/**
		 * Write the labels of the training samples to a label file.
		 * @param[in]	modelInfo	information of the binary classification model
		 * @return		true if success; false if failure
		 */
		bool writeLabelFile(const BinaryModel &modelInfo);

		/**
		 * Read the results of binary classification from a label file.
		 * @param[in]	modelName	model name
		 * @param[out]	results		binary classification results
		 * @return		true if success; false if failure
		 */
		bool readLabelFile(const std::string &modelName, std::vector<bool> &results);

		/**
		 * Collect the class information from the training samples.
		 * @param[in]	labels		labels of the training samples
		 * @return		number of classes
		 */
		unsigned int collectClassInfo(const std::vector<int> &labels);

		/**
		 * Vote according to the one vs. one classification results.
		 * @param[in]	votes		results collected from the one vs. one scheme
		 * @param[out]	results		predicted class labels
		 */
		void vote(const UIntMatrix &votes, std::vector<int> &results);

		/**
		 * Delete a file.
		 * @param[in]	fileName	base file name
		 * @param[in]	fileExt		file extension
		 * @return		true if success; false if failure
		 */
		static bool _deleteFile(const std::string &fileName, const std::string &fileExt);

		/**
		 * Clear a directory.
		 * Delete all the files under a directory.
		 * @param[in]	directoryName	path of the directory
		 * @return		true if success; false if failure
		 */
		static bool _clearDirectory(const std::string &directoryName);

	private:
		ParaType parameters;								/**< Parameters for the PSVM classifier. */
		std::vector< std::vector<unsigned int> > classMap;	/**< Map for class members. */
		std::vector<int> labelMap;							/**< Map for class labels. */
		std::vector<BinaryModel> modelMap;					/**< Map for binary classification models. */
	};
}

#endif // IDL_PSVM_H
