// File: RegLocalSDA.cpp
// Purpose: Implementation of the RegLocalSDA class
// Remarks: This class is designed for RegLocalSDA classification.
// Author: Peter Sadowski peterjsadowski@gmail.com

#include <cstdlib>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "RegLocalSDA.h"
#include "LocalSDAModel.h"
#include "math.h"
#define INF 1.7E308;

using namespace std;

namespace idl
{
	// Compute "global" sim pmfs and prior.
	bool RegLocalSDA::configure(const SimData &data)
	{
		if (parameters.neighborhoodSize == 0)
			return false;
		else if(parameters.neighborhoodSize >= data.getTrainingSetSize()) {
			cout << "Error: Neighborhood size too big." << endl;
			return false; }

		RealMatrix simDomain;
		vector<int> uniqueLabels;
		vector<int> trainingLabels;
		data.getSimDomain(simDomain);
		data.getUniqueLabels(uniqueLabels);
		data.getTrainingLabels(trainingLabels);
		unsigned int nTraining = data.getTrainingSetSize();

		// Create map structure to hold mean sim probabilities for each (g,h)
		// and map structure to hold mean class label priors
		for (vector<int>::const_iterator itrG = uniqueLabels.begin();
				itrG != uniqueLabels.end(); ++itrG)
		{
			regLabelPriors[*itrG] = 0;
		}
		// Map similarities to their indices in simDomain.
		for(unsigned int i = 0; i < simDomain.size()[0]; ++i) 
		{
			simIdx[simDomain(i,0)] = i;
		}

		// Initialize some data structures.
		// combinedHists will map (G,H) label pair to histogram of observed sim values over all local neighborhoods.
		map<vector<int>,RealMatrix> combinedHists;
		map<vector<int>,RealMatrix> PMF_Sum_ICML09;
		for (vector<int>::const_iterator itrG = uniqueLabels.begin();
				itrG != uniqueLabels.end(); ++itrG)
		{
			for (vector<int>::const_iterator itrH = uniqueLabels.begin();
				itrH != uniqueLabels.end(); ++itrH)
			{
				vector<int> labelKeyGH(2);
				labelKeyGH[0] = *itrG;
				labelKeyGH[1] = *itrH;
				// For calculating PAvg_Peter;
				combinedHists[labelKeyGH].zeros(simDomain.size()[0], simDomain.size()[1]);
				// For calculating PAvg_ICML09;
				PMF_Sum_ICML09[labelKeyGH].zeros(simDomain.size()[0], simDomain.size()[1]);
			}
		}

		// Calculate regularizing probabilities for each training sample, label g, label h
		// and also mean class label priors for each label.	
		for (unsigned int i = 0; i < nTraining; ++i) // Iterate over each training samples.
		{
			LocalSDAModel lSDA;
			RealMatrix simMat(parameters.neighborhoodSize, parameters.neighborhoodSize);
			vector<int> labels(parameters.neighborhoodSize);
			RealMatrix simVec(1, parameters.neighborhoodSize);
			data.getNNofTrainingSimVec(i, parameters.neighborhoodSize, simVec);
			data.getNNofTrainingLabels(i, parameters.neighborhoodSize, labels);
			data.getNNofTrainingSimMat(i, parameters.neighborhoodSize, simMat);
			lSDA.LocalSDAModel::initialize(simMat, labels, simVec, simDomain);

			regLabelPriors[trainingLabels[i]] += 1; // Update class of training sample prior.

			for (vector<int>::const_iterator itrG = uniqueLabels.begin(); itrG != uniqueLabels.end(); ++itrG)
			{
				for (vector<int>::const_iterator itrH = uniqueLabels.begin(); itrH != uniqueLabels.end(); ++itrH)
				{
					RealMatrix HCenterToGSims;
					lSDA.getHCenterToGSims(simMat,HCenterToGSims,*itrH, *itrG);

					if (!HCenterToGSims.isEmpty())
					{
						vector<int> labelKeyGH(2);
						labelKeyGH[0] = *itrG;
						labelKeyGH[1] = *itrH;
						RealMatrix localHist;	// For calculating PMF_Sum_ICML09.
						localHist.zeros(simDomain.size()[0], simDomain.size()[1]);
						// Update combined hist data structure.
						for (unsigned int j = 0; j < HCenterToGSims.length(); ++j)
						{
							combinedHists[labelKeyGH](simIdx[HCenterToGSims(j)]) += 1;
							localHist(simIdx[HCenterToGSims(j)]) += 1;
						}
						// Update structure for calculating ICML09 average-of-local-discrete-exponentials.
						assert(localHist.sum() > 0);
						localHist /= localHist.sum();	// Get PMF.
						localHist = idl::getDiscreteExponential(localHist, simDomain); // Get discrete exponential PMF.
						PMF_Sum_ICML09[labelKeyGH] += localHist;					   // Add to the sum of PMFs.
					}			
				}
			}
		}

		// Perform laplace smoothing on mean class label priors
		for (vector<int>::const_iterator itrG = uniqueLabels.begin();
				itrG != uniqueLabels.end(); ++itrG)
		{
			regLabelPriors[*itrG] = (regLabelPriors[*itrG] + 1)/(nTraining + uniqueLabels.size());

			for (vector<int>::const_iterator itrH = uniqueLabels.begin();
				itrH != uniqueLabels.end(); ++itrH)
			{
				vector<int> labelKeyGH(2);
				labelKeyGH[0] = *itrG;
				labelKeyGH[1] = *itrH;
				// Compute PAvg_Peter.
				if (combinedHists[labelKeyGH].sum() > 0) {
					// Normalize to get PMF.
					PAvg_Peter[labelKeyGH] = combinedHists[labelKeyGH] / combinedHists[labelKeyGH].sum();
					// Perform discrete exponential smoothing of PMF.
					PAvg_Peter[labelKeyGH] = idl::getDiscreteExponential(PAvg_Peter[labelKeyGH], simDomain);
				} else {
					// Use uniform distribution.
					// See note below.
					PAvg_Peter[labelKeyGH].ones(simDomain.size()[0], 1);
					PAvg_Peter[labelKeyGH] = PAvg_Peter[labelKeyGH] / PAvg_Peter[labelKeyGH].sum();
				}
				// Compute PAvg_ICML09.
				if (PMF_Sum_ICML09[labelKeyGH].sum() > 0) {
					PAvg_ICML09[labelKeyGH] = PMF_Sum_ICML09[labelKeyGH] / PMF_Sum_ICML09[labelKeyGH].sum();
				} else {
					// Use uniform distribution.
					// Note: Peter and Luca discussed this and decided that a uniform distribution would be
					//	     a reasonable solution to this situation.
					PAvg_ICML09[labelKeyGH].ones(simDomain.size()[0], 1);
					PAvg_ICML09[labelKeyGH] = PAvg_ICML09[labelKeyGH] / PAvg_ICML09[labelKeyGH].sum();
				}

			}
		}
		
		return true;
	}

	// Classify the test samples of the data set, and get their class probabilities.
	bool RegLocalSDA::classify(const SimData &data, vector<int> &classPredictions, RealMatrix &classProbMatrix)
	{
		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size.
		if (data.getTrainingSetSize() < parameters.neighborhoodSize)
		{
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}
		// Initialize data structures.
		unsigned int numTestSamples = data.getTestSetSize();
		RealMatrix simDomain;		// Possible sim values over all data.
		data.getSimDomain(simDomain);
		vector<int> uniqueLabels;	// Unique labels over all data.
		data.getUniqueLabels(uniqueLabels);
		classPredictions.resize(numTestSamples);
		unsigned int nUniqueLabels = (unsigned int)uniqueLabels.size();
		classProbMatrix.resize(numTestSamples,nUniqueLabels);	// Class probabilities for each test point.			
		// For each test sample, classify and get class probabilities.
		for (unsigned int i = 0; i < numTestSamples; ++i)
		{
			// Initialize neighborhood-specific data structures.
			RealMatrix simMat(parameters.neighborhoodSize,
						parameters.neighborhoodSize);				// Local sim matrix of neighborhood.
			vector<int> labels(parameters.neighborhoodSize);		// Labels of neighborhood.
			RealMatrix simVec(1, parameters.neighborhoodSize);		// Sim of test point to neighborhood.
			data.getNNSimMat(i, parameters.neighborhoodSize, simMat);
			data.getNNLabels(i, parameters.neighborhoodSize, labels);
			data.getNNSimVec(i, parameters.neighborhoodSize, simVec);
			LocalSDAModel unregLsdaModel;
			unregLsdaModel.initialize(simMat, labels, simVec, simDomain);
			vector<int> uniqueLocalLabels;				// List of unique labels in this neighborhood.
			unregLsdaModel.getUniqueLabels(uniqueLocalLabels);
			// Maybe do this only if necessary? (handle degenerate cases)
			if (unregLsdaModel.haveReasonToUseBackupClassifier()) {
				// We want to use the backup classifier in LocalSDAModel.
				int predictedClass;
				map<int,double> tempProbs;
				if (!unregLsdaModel.backupClassify(simMat, labels, simVec, simDomain, predictedClass, tempProbs)) {
						return false;}
				classPredictions[i] = predictedClass;
				// Convert prob map to matrix form.
				for (unsigned int labelIdx = 0; labelIdx < uniqueLabels.size(); ++labelIdx)
				{
					if (tempProbs.count(uniqueLabels[labelIdx]))
					{
						classProbMatrix(i,labelIdx) = tempProbs[uniqueLabels[labelIdx]];
					}
				}
				continue;	// Continue to next sample.
			}
			// Compute full Local SDA model.
			unregLsdaModel.computeModelParams(simMat, labels, simVec, simDomain);  
			RealMatrix classProbsTest_i;				// Holds the class probabilities of testpoint i.
			classProbsTest_i.zeros(1,nUniqueLabels);
			double maxProb = (-1)*INF;	
			int maxProbLabel = 0;
			// Iterate over neighborhood class labels G.
			for (vector<int>::const_iterator itrG = uniqueLocalLabels.begin(); 
				itrG != uniqueLocalLabels.end(); ++itrG)
			{
				// Calculate regularized likelihood of testpoint being in class G.
				double tmpRegProb;
				if (parameters.jointFlag)
					tmpRegProb = regJointProbability(unregLsdaModel,*itrG, simVec, simDomain);
				else if(parameters.marginalFlag)
					tmpRegProb = regMarginalProbability(unregLsdaModel,*itrG, simVec, simDomain);
				else if (parameters.peterFlag)
					tmpRegProb = regPeterProbability(unregLsdaModel,*itrG, simVec, simDomain);
				else {
					cerr << "No regularization flag set." << endl;
					return false;
				}
				// Keep track of the most likely label.
				if (tmpRegProb > maxProb)
				{
					maxProb = tmpRegProb;
					maxProbLabel = *itrG;
				}
				// Save likelihood of this label G to a vector.
				for (unsigned int j = 0; j < nUniqueLabels; ++j)
				{
					if (uniqueLabels[j] == *itrG)
					{
						// j is the index of label *itrG. 
						classProbsTest_i(j) = tmpRegProb;
					}
				}
			}  // End iteration over G.
			// Normalize likelihoods to get probabilities.
			classProbsTest_i /= classProbsTest_i.sum();
			// Record classPredictions and probabilities.
			for (unsigned int j = 0; j < nUniqueLabels; j++)
				classProbMatrix(i,j) = classProbsTest_i(j);
			classPredictions[i] = maxProbLabel;
			// Check that probabilities sum to 1.
			if (classProbMatrix.sumRow(i) < .9999 || classProbMatrix.sumRow(i) > 1.0001)
			{		
				cerr << "Probs not summing to 1 in RegLocalSDA classifier." << endl;
				return false;
			}
		} // End iteration over test points.
		return true;
	}

	// Classify the test samples of the data set.
	bool RegLocalSDA::classify(const SimData &data, vector<int> &classPredictions)
	{
		RealMatrix classProbMatrix;
		return classify(data,classPredictions,classProbMatrix);
	}
	
	// Marginal regularization method.
	double RegLocalSDA::regMarginalProbability(LocalSDAModel &lsda, int g, RealMatrix &testSimVec, const RealMatrix &simDomain)
	{
		bool debug = false;

		vector<int> uniqueLocalLabels;
		lsda.getUniqueLabels(uniqueLocalLabels);

		double Pg = 1;
		
		if (debug)
			std::cout << "unreg P \tregTerm\n";

		for (vector<int>::const_iterator itrH = uniqueLocalLabels.begin(); 
			itrH != uniqueLocalLabels.end(); ++itrH)
		{
			vector<int> keyGH(2);
			keyGH[0] = g;
			keyGH[1] = *itrH;
			RealMatrix tmpReg = PAvg_ICML09[keyGH];

			// unregularized probability
			//   P(s(z,u_h) = s' | Y = g) 
			double Phg = (1-parameters.regParam)*lsda.getSimProbGivenClass(testSimVec,g,*itrH, simDomain);
			
			// DEBUG
			if (debug)
				std::cout << lsda.getSimProbGivenClass(testSimVec,g,*itrH, simDomain) << "\t";

			// regularized probability 
			//    P_hat(s(z,u_h) = s' | Y = g)
			// regularization term:
			//    sum ( P(x,u_h) = s' | Y = g ); x in training samples
			double simTestToH = lsda.getSimTestToH(testSimVec,*itrH);
			Phg +=	parameters.regParam*tmpReg(simIdx[simTestToH]);

			// DEBUG
			if (debug)
				std::cout << tmpReg(simIdx[simTestToH]) << "\t";
		
			// naive bayes produt of marginals
			//     P_hat = product ( P_hat(s(z,u_h) = s' | Y = g) ) over h in local centroids H
			Pg *= Phg;
		}
		// multiply by regularized class prior of class G
		// PJS: THIS SEEMS WRONG
		Pg *= ((1-parameters.regParam) * lsda.getPrior(g) + (parameters.regParam) * regLabelPriors[g]);
		//Pg *= ((1-regParam) * lsda.getPrior(g) + (regParam) * g)*g;

		if (debug) {
			std::cout << std::endl << "unregPrior \tregPrior\n";
			std::cout << lsda.getPrior(g) << "\t\t" << regLabelPriors[g] <<std::endl;
			std::cout << "prob: " << Pg << std::endl <<std::endl;
		}
		return Pg;
	}
	
	double RegLocalSDA::regJointProbability(LocalSDAModel &lsda, int g, RealMatrix &testSimVec, const RealMatrix &simDomain)
	{
		vector<int> uniqueLocalLabels;
		lsda.getUniqueLabels(uniqueLocalLabels);

		double nonRegProb = 1;
		double regularizingTerm = 1;
		for (vector<int>::const_iterator itrH = uniqueLocalLabels.begin(); 
			itrH != uniqueLocalLabels.end(); ++itrH)
		{
			vector<int> keyGH(2);
			keyGH[0] = g;
			keyGH[1] = *itrH;
			RealMatrix tmpReg = PAvg_ICML09[keyGH];
			// joint regularization term:
			//    sum ( P(x,u_h) = s' | Y = g ); x in training samples
			double simTestToH = lsda.getSimTestToH(testSimVec,*itrH);
			regularizingTerm *=	tmpReg(simIdx[simTestToH]);

			// joint unregularized probability
			//   P(s(z,u_h) = s' | Y = g) 
			nonRegProb *= lsda.getSimProbGivenClass(testSimVec,g,*itrH, simDomain);

			// naive bayes produt of marginals
			//     P_hat = product ( P_hat(s(z,u_h) = s' | Y = g) ) over h in local centroids H
		}

		// add joint regularizing term to unregularized joint probability
		return (1-parameters.regParam) * lsda.getPrior(g) * nonRegProb + (parameters.regParam) * regLabelPriors[g] * regularizingTerm;
	}



	void RegLocalSDA::regMarginalJointProbability(LocalSDAModel &lsda, int g, RealMatrix &testSimVec, const RealMatrix &simDomain, 
												  vector<double> &probabilities) 
	{	
		vector<int> uniqueLocalLabels;
		lsda.getUniqueLabels(uniqueLocalLabels);

		double PgMarg = 1;
		double nonRegProb = 1;
		double regularizingTerm = 1;
		for (vector<int>::const_iterator itrH = uniqueLocalLabels.begin(); 
			itrH != uniqueLocalLabels.end(); ++itrH)
		{
			vector<int> keyGH(2);
			keyGH[0] = g;
			keyGH[1] = *itrH;
			RealMatrix tmpReg = PAvg_ICML09[keyGH];

			// Unregularized probability.
			//   P(s(z,u_h) = s' | Y = g) 
			double tmpNonRegProb = lsda.getSimProbGivenClass(testSimVec,g,*itrH, simDomain);
			double Phg = (1-parameters.regParam)*tmpNonRegProb;
			nonRegProb *= tmpNonRegProb;

			// regularized probability 
			//    P_hat(s(z,u_h) = s' | Y = g)
			// regularization term:
			//    sum ( P(x,u_h) = s' | Y = g ); x in training samples
			double simTestToH = lsda.getSimTestToH(testSimVec,*itrH);
			double tmpRegTerm = tmpReg(simIdx[simTestToH]);
			Phg +=	parameters.regParam*tmpRegTerm;
			regularizingTerm *=	tmpRegTerm;

			// naive bayes produt of marginals
			//     P_hat = product ( P_hat(s(z,u_h) = s' | Y = g) ) over h in local centroids H
			PgMarg *= Phg;
		}
		// multiply by regularized class prior of class G
		double unRegPrior = lsda.getPrior(g);
		double regPrior = regLabelPriors[g];
		PgMarg *= ((1-parameters.regParam) * unRegPrior + (parameters.regParam) * regPrior);//*itrG);
		double PgJoint = (1-parameters.regParam) * unRegPrior * nonRegProb + (parameters.regParam) * regPrior * regularizingTerm;
	
		probabilities.resize(2);
		probabilities[0] = PgMarg;
		probabilities[1] = PgJoint;
	}

	double RegLocalSDA::regPeterProbability(LocalSDAModel &lsda, int g, RealMatrix &testSimVec, const RealMatrix &simDomain)
	{
		vector<int> uniqueLocalLabels;
		lsda.getUniqueLabels(uniqueLocalLabels);

		double combinedLikelihood = 1;
		for (vector<int>::const_iterator itrH = uniqueLocalLabels.begin(); 
			itrH != uniqueLocalLabels.end(); ++itrH)
		{
			vector<int> keyGH(2);
			keyGH[0] = g;
			keyGH[1] = *itrH;
			// Similarity of test point to H centroid.
			double simTestToH = lsda.getSimTestToH(testSimVec,*itrH);
			// Regularized likelihood:
			// P( s(x,u_h) = s' | Y = g ); x in training samples
			RealMatrix regPMF = PAvg_Peter[keyGH];
			double regLikelihood = regPMF(simIdx[simTestToH]);

			//  "Unregularized" likelihood based on neighborhood.
			//   P(s(z,u_h) = s' | Y = g); z in local neighborhood.
			double nonRegLikelihood = lsda.getSimProbGivenClass(testSimVec,g,*itrH, simDomain);

			// naive bayes produt of marginals
			//     P_hat = product ( P_hat(s(z,u_h) = s' | Y = g) ) over h in local centroids H
			combinedLikelihood *= (1-parameters.regParam) * nonRegLikelihood + parameters.regParam * regLikelihood;
		}

		// Posterior Prob is proportional to Likelihood * Prior;
		double posterior = combinedLikelihood * lsda.getPrior(g);
		assert(posterior >= 0 && posterior < 1);
		return posterior;
	}
}


