// File: RegLocalSDATest.cpp
// Purpose: Implementation of the RegLocalSDATest class
// Remarks: This class is designed as an adapter for the WeightedKNN class
//			to enable testing on multiple sets of parameters.
// Author: Ingrid Liu (ingrid@hkn.eecs.berkeley.edu)
// Last modified: 6/20/2009 4:24:41 PM

#include <cmath>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "RegLocalSDATest.h"
#include "LocalSDATest.h"

using namespace std;

namespace idl
{
	const string RegLocalSDATest::_ID = "rlSDA"; // Define the class ID

	// Register the class instantiator with the factory
	const bool RegLocalSDATest::_REGISTERED =
		ClassifierTestFactory::registerCreator(RegLocalSDATest::_ID, &RegLocalSDATest::create);

	// Destructor
	RegLocalSDATest::~RegLocalSDATest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Configure the test for a weighted k-NN classifier
	bool RegLocalSDATest::configure(ConfigParser &config, int secIdx)
	{
		std::cout << "configuring... Reg\n";
		string strRegularizedMode;

		//Get regularization mode for regularized lSDA
		//marginal, joint, none
		if (!config.getString(secIdx, "RegularizedMode", strRegularizedMode))
		{
			cerr << "Missing regularization mode for rlSDA classifier " << secIdx << endl;
			return false;
		}
		
		setRegularizationFlags(strRegularizedMode, regularizeFlag, marginalFlag, jointFlag, bothFlag, peterFlag);

		if (strRegularizedMode != "none")
		{
			// Get regularization parameters to cross-validate
			if (!config.getRealVec(secIdx, "regParam", regParams))
			{
				cerr << "Error: Missing or incorrect regParam for rlSDA classifier "
					 << secIdx << endl;
				return false;
			}
		}
		else
		{
			regParams.clear();
			regParams.push_back(0.0);
		}
		
		// Get neighborhood sizes to cross-validate
		if (!config.getUIntVec(secIdx, "NeighborhoodSize", neighborhoodSizes))
		{
			cerr << "Error: Missing or incorrect neighborhood size for kNN classifier "
				 << secIdx << endl;
			return false;
		}

		//Check the retrieved neighborhood sizes
		if (!checkNeighborhoodSizes())
		{
			cerr << "Error: Incorrect neighborhood size for rlSDAclassifier "
				 << secIdx << endl;
			return false;
		}

		reset(static_cast<unsigned int>(regParams.size() * neighborhoodSizes.size()));

		return true;
	}

	// Test the weighted k-NN classifier on all available parameters
	bool RegLocalSDATest::test(const SimData &data, ofstream &logFile)
	{
		// Log the weight mode
		if (! regularizeFlag) 
		{
			// non-regularized mode simplifies to LocalSDATest
			LocalSDATest lSDATest(neighborhoodSizes);
			lSDATest.test(data, logFile);
		}
		else {
			string strMode = "";
			string errorTab = "";
			if (marginalFlag) 
				strMode = "Marginal";
			else if (jointFlag) 
				strMode = "Joint";
			else if (peterFlag) 
				strMode = "Peter";
			else
				cerr << "Regularization Scheme not defined" << endl;

			logFile << strMode << " Regularized Local SDA classifier" << endl
					<< "Neighborhood size\tRegParam\tError Rate" << endl;

			unsigned int i = 0; // parameter set count
			vector<int> results(data.getTestSetSize());

			for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
				 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
			{
				RegLocalSDA rlsda;
				rlsda.setNeighborhoodSize(*itrNeighborhoodSize);
				rlsda.setFlag(marginalFlag,jointFlag, peterFlag);
				
				// calculate regularizizing probabilities
				if (!rlsda.configure(data)) return false;

				for (vector<double>::const_iterator itrRegParam = regParams.begin();
					 itrRegParam != regParams.end(); ++itrRegParam)
				{
					rlsda.setRegPara(*itrRegParam);

					// classify all test points
					
					if (!rlsda.classify(data, results)) return false;

					double errorRate = data.getClassificationError(results);
				
					// Log the error rate
					logFile << *itrNeighborhoodSize << "\t" << *itrRegParam << "\t"
							<< (errorRate * 100.0) << "%" << endl;

					errorSum(i) += errorRate;
					errorSumOfSquares(i++) += (errorRate * errorRate);
				}
			}
		}
		logFile << endl;

		++testCount;

		return true;
	}

	bool RegLocalSDATest::setRegularizationFlags(string regMode, bool &regularizeFlag, bool &marginalFlag, bool &jointFlag, bool &bothFlag, bool &peterFlag)
	{
		regularizeFlag = true;
		marginalFlag = false;
		jointFlag = false;
		bothFlag = false;
		peterFlag = false;

		if (regMode == "none")
			regularizeFlag = false;
		else if (regMode == "marginal")
			marginalFlag = true;
		else if (regMode == "joint")
			jointFlag = true;
		else if (regMode == "peter")
			peterFlag = true;
		else
			return false;
		return true;
	}

	// Check the legitimacy of the neighborhood sizes
	bool RegLocalSDATest::checkNeighborhoodSizes()
	{
		sort(neighborhoodSizes.begin(), neighborhoodSizes.end());

		return neighborhoodSizes.front() > 0;
	}

	// Clear the results with the best parameters
	void RegLocalSDATest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void RegLocalSDATest::selectBestPara()
	{
		
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet

		if (0 == testCount)
		{
			// No cross-validation performed
			latestResult.regParam = regParams.front();
			latestResult.neighborhoodSize = neighborhoodSizes.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);

			// Get the neighborhood size corresponding to the minimum cross-validation error
			latestResult.neighborhoodSize = neighborhoodSizes[i / regParams.size()];

			// Get the regParam corresponding to the minimum cross-validation error
			latestResult.regParam = regParams[i % regParams.size()];
		}

		resultsBestPara.push_back(latestResult);
	}

	// Test the weighted k-NN classifier with the best parameters
	bool RegLocalSDATest::testBestPara(const SimData &data, ofstream &logFile)
	{
		string regMode = "";
		if (marginalFlag)
			regMode = "Marginal";
		else if (jointFlag)
			regMode = "Joint";
		else if (peterFlag)
			regMode = "Peter";
		else
			return false;	

		logFile << regMode << " Regularized Local SDA classifier" << endl;

		RegLocalSDA rlsda;
		unsigned int k = (resultsBestPara.back()).neighborhoodSize;
		rlsda.setNeighborhoodSize(k);
		rlsda.setFlag(marginalFlag,jointFlag,peterFlag);
		double r = (resultsBestPara.back()).regParam;
		rlsda.setRegPara(r);
		

		rlsda.configure(data);

		vector<int> results(data.getTestSetSize());
		
		RealMatrix probMatrix;
		if (!rlsda.classify(data, results, probMatrix)) return false;

		logFile << "Classification probabilities:" << endl;
		for (unsigned int i = 0; i < probMatrix.getRowNum(); ++i)
		{
			for (unsigned int j = 0; j < probMatrix.getColumnNum(); ++j)
			{
				logFile << probMatrix(i,j) << " ";
			}
			logFile << endl;
		}
		logFile << "End of classification probabilities." << endl;

		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;

		// Log the error rate
		logFile << "Neighborhood size\tLambda\tError rate" << endl
				<< k << "\t" << (errorRate * 100.0) << "%" << endl << endl;

		return true;
	}

	// Report the test results of all available parameters
	void RegLocalSDATest::reportResults(ofstream &file, bool verbose)
	{
		string regMode = "";
		if (marginalFlag)
			regMode = "Marginal";
		else if (jointFlag)
			regMode = "Joint";
		else if (peterFlag)
			regMode = "Peter";
			
		file << regMode << " Regularized Local SDA classifier" << endl;

		if (verbose)
		{
			// Report the results of all the parameter sets
			file << "Neighborhood size\tRegParam\tError rate\tStandard deviation" << endl;

			unsigned int i = 0;
			for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
				 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
			{
				for (vector<double>::const_iterator itrRegParam = regParams.begin();
					 itrRegParam != regParams.end(); ++itrRegParam)
				{
					file << *itrNeighborhoodSize << "\t" << *itrRegParam << "\t"
						 << (getError(i) * 100.0) << "%\t" << (getErrorStdDev(i) * 100.0) << "%" << endl;
					++i;
				}
			}
		}
		else
		{
			// Report the best result only
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "RegParam: " << (resultsBestPara.back()).regParam << endl
				 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Report the test results with the best parameters
	void RegLocalSDATest::reportBestParaResults(ofstream &file, bool verbose)
	{
		string regMode = "";
		if (marginalFlag)
			regMode = "Marginal";
		else if (jointFlag)
			regMode = "Joint";
		else if (peterFlag)
			regMode = "Peter";
			
		file << regMode << " Regularized Local SDA classifier" << endl;

		if (verbose)
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "RegParam: " << (resultsBestPara.back()).regParam << endl
				 << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}
		else
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "RegParam: " << (resultsBestPara.back()).regParam << endl
				 << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Summarize the results of all the tests with the best parameters
	void RegLocalSDATest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		string regMode = "";
		if (marginalFlag)
			regMode = "Marginal";
		else if (jointFlag)
			regMode = "Joint";
		else if (peterFlag)
			regMode = "Peter";
			
		file << regMode << " Regularized local SDA classifier" << endl;

		if (verbose)
		{
			file << "Neighborhood size\tRegParam\tError rate" << endl;
		}

		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		RealMatrix kVec(1, n);
		RealMatrix regParamVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			kVec(i) = static_cast<double>(itr->neighborhoodSize);
			regParamVec(i) = itr->regParam;
			errorVec(i) = itr->testError;

			if (verbose)
			{
				file << itr->neighborhoodSize << "\t" << itr->regParam << "\t"
					 << (itr->testError * 100.0) << "%" << endl;
			}
		}
		
		file << "Average neighborhood size: " << kVec.mean()
			 << " (" << kVec.std() << ")" << endl
			 << "Average regParam: " << regParamVec.mean()
			 << " (" << regParamVec.std() << ")" << endl
			 << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
