// File: SVMKNNTest.h
// Purpose: Definition of the SVMKNNTest class
// Remarks: This class is designed as an adapter for the SVMKNN class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/27/2008 4:21:09 PM

#ifndef IDL_SVMKNNTEST_H
#define IDL_SVMKNNTEST_H

#include <vector>
#include <string>
#include <fstream>
#include "ClassifierTest.h"
#include "SVMKNN.h"

namespace idl
{
	/**
	 * Class for testing an SVM-KNN classifier on similarity data.
	 * An adapter class for the SVMKNN class to enable testing on multiple
	 * sets of parameters.
	 * @remarks	This class is derived from the ClassifierTest class.
	 * @see		ClassifierTest
	 * @see		ClassifierTestFactory
	 * @see		SVMKNN
	 */
	class SVMKNNTest : public ClassifierTest
	{
	public:
		SVMKNNTest() : ClassifierTest() {}	/**< Constructor. */
		~SVMKNNTest();						/**< Destructor. */

		/** Clear the results with the best parameters. */
		void resetBestParaResults();

		/** Select the best parameters based on the previous tests. */
		void selectBestPara();

		/**
		 * Configure the test for an SVM-KNN classifier.
		 * @param[in]	config		ConfigParser object containing user configurations
		 * @param[in]	secIdx		section index
		 * @return		true if success; false if failure
		 */
		bool configure(ConfigParser &config, int secIdx);

		/**
		 * Test the SVM-KNN classifier on all available parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool test(const SimData &data, std::ofstream &logFile);

		/**
		 * Test the SVM-KNN classifier with the best parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool testBestPara(const SimData &data, std::ofstream &logFile);

		/**
		 * Report the test results of all available parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 * @remarks		For concise output, only the best result and its corresponding
		 *				parameters are reported. For verbose output, the results of all
		 *				available parameter sets are reported.
		 */
		void reportResults(std::ofstream &file, bool verbose = false);

		/**
		 * Report the test results with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaResults(std::ofstream &file, bool verbose = false);

		/**
		 * Summarize the results of all the tests with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaSummary(std::ofstream &file, bool verbose = false);

	private:
		/**
		 * Check the legitimacy of the penalty parameters for C-SVM.
		 * @return	true if all the penalty parameters are within the legal range;
		 *			false if at least one penalty parameter is out of the legal range
		 */
		bool checkCs();

		/**
		 * Check the legitimacy of the neighborhood sizes.
		 * @return	true if all the neighborhood sizes are within the legal range;
		 *			false if at least one neighborhood size is out of the legal range
		 */
		bool checkNeighborhoodSizes();

	private:
		SVMKNN svmknn; /**< SVM-KNN classifier. */
		std::vector<double> Cs; /**< Vector of penalty parameters for C-SVM. */
		std::vector<unsigned int> neighborhoodSizes; /**< Vector of neighborhood sizes. */

		/** Structure type for test results with the best parameters. */
		struct ResultType
		{
			unsigned int neighborhoodSize;	/**< Neighborhood size. */
			double C;						/**< Penalty parameter for C-SVM. */
			double cvError;					/**< Cross-validation error. */
			double cvErrorStdDev;			/**< Standard deviation of the cross-validation error. */
			double testError;				/**< Test error. */
		};

		std::vector<ResultType> resultsBestPara; /**< Test results with the best parameters. */

	private:
		/**
		 * Create an object of the SVMKNNTest class.
		 * @return	pointer to a new object of the SVMKNNTest class
		 * @remarks	This is the creator function to be registered with the factory.
		 */
		static ClassifierTest *create()
		{
			return new SVMKNNTest;
		}
		static const std::string _ID;	/**< Class ID ("SVM-KNN") */
		static const bool _REGISTERED;	/**< Flag for creator registration */
	};
}

#endif // IDL_SVMKNNTEST_H
