// File: SVMTest.cpp
// Purpose: Implementation of the SVMTest class
// Remarks: This class is designed as an adapter for the SVM class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/27/2008 4:23:56 PM

#include <iostream>
#include <algorithm>
#include "SVMTest.h"

using namespace std;

namespace idl
{
	const string SVMTest::_ID = "SVM"; // Define the class ID

	// Register the class instantiator with the factory
	const bool SVMTest::_REGISTERED = 
		ClassifierTestFactory::registerCreator(SVMTest::_ID, &SVMTest::create);

	// Destructor
	SVMTest::~SVMTest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Sort the elements of a real vector and check their positivity
	bool SVMTest::checkPositive(vector<double> &vec)
	{
		sort(vec.begin(), vec.end());

		return vec.front() > 0.0;
	}

	// Configure the test for an SVM classifier
	bool SVMTest::configure(ConfigParser &config, int secIdx)
	{
		// Get the flag for strictly supervised learning
		string strStrictlySupervised;
		if (!config.getString(secIdx, "StrictlySupervised", strStrictlySupervised))
		{
			cerr << "Error: Missing flag for strictly supervised learning for SVM classifier "
				 << secIdx << endl;
			return false;
		}
		if ("yes" == strStrictlySupervised)
		{
			svm.setStrictlySupervisedFlag(true);
		}
		else if ("no" == strStrictlySupervised)
		{
			svm.setStrictlySupervisedFlag(false);
		}
		else
		{
			cerr << "Error: Unknown flag for strictly supervised learning for SVM classifier "
				 << secIdx << endl;
			return false;
		}

		// Get the flag for compensated projection
		string strCompensatedProjection;
		if (!config.getString(secIdx, "CompensatedProjection", strCompensatedProjection))
		{
			cerr << "Error: Missing flag for compensated projection for SVM classifier "
				 << secIdx << endl;
			return false;
		}
		if ("yes" == strCompensatedProjection)
		{
			svm.setCompensatedProjectionFlag(true);
		}
		else if ("no" == strCompensatedProjection)
		{
			svm.setCompensatedProjectionFlag(false);
		}
		else
		{
			cerr << "Error: Unknown flag for compensated projection for SVM classifier "
				 << secIdx << endl;
			return false;
		}

		// Get the kernel type
		string strKernel;
		if (!config.getString(secIdx, "Kernel", strKernel))
		{
			cerr << "Error: Missing kernel type for SVM classifier " << secIdx << endl;
			return false;
		}
		if ("linear" == strKernel)			// Linear kernel
		{
			svm.setKernelType(SVM::KERNEL_LINEAR);
		}
		else if ("poly" == strKernel)		// Polynomial kernel
		{
			svm.setKernelType(SVM::KERNEL_POLY);
		}
		else if ("rbf" == strKernel)		// RBF kernel
		{
			svm.setKernelType(SVM::KERNEL_RBF);
		}
		else if ("sigmoid" == strKernel)	// Sigmoid kernel
		{
			svm.setKernelType(SVM::KERNEL_SIGMOID);
		}
		else if ("sim_shift" == strKernel)	// Similarity kernel with eigenvalue shift
		{
			svm.setKernelType(SVM::KERNEL_SIM_SHIFT);
		}
		else if ("sim_clip" == strKernel)	// Similarity kernel with eigenvalue clip
		{
			svm.setKernelType(SVM::KERNEL_SIM_CLIP);
		}
		else if ("sim_flip" == strKernel)	// Similarity kernel with eigenvalue flip
		{
			svm.setKernelType(SVM::KERNEL_SIM_FLIP);
		}
		else
		{
			cerr << "Error: Unknown kernel type for SVM classifier " << secIdx << endl;
			return false;
		}

		// Get the kernel parameter "gamma" for RBF, polynomial or sigmoid kernel
		if ("rbf" == strKernel || "poly" == strKernel || "sigmoid" == strKernel)
		{
			if (!config.getRealVec(secIdx, "gamma", gammas))
			{
				cerr << "Error: Missing or incorrect gamma for SVM classifier " << secIdx << endl;
				return false;
			}
			if (!checkPositive(gammas))
			{
				cerr << "Error: Incorrect gamma for SVM classifier " << secIdx << endl;
				return false;
			}
		}
		else
		{
			gammas.clear();
			gammas.push_back(0.0);
		}

		// Get the kernel parameter "r" for polynomial or sigmoid kernel
		if ("poly" == strKernel || "sigmoid" == strKernel)
		{
			if (!config.getRealVec(secIdx, "r", rs))
			{
				cerr << "Error: Missing or incorrect r for SVM classifier " << secIdx << endl;
				return false;
			}
			if (!checkPositive(rs))
			{
				cerr << "Error: Incorrect r for SVM classifier " << secIdx << endl;
				return false;
			}
		}
		else
		{
			rs.clear();
			rs.push_back(0.0);
		}

		// Get the kernel parameter "d" for polynomial kernel
		if ("poly" == strKernel)
		{
			if (!config.getRealVec(secIdx, "d", ds))
			{
				cerr << "Error: Missing or incorrect d for SVM classifier " << secIdx << endl;
				return false;
			}
			if (!checkPositive(ds))
			{
				cerr << "Error: Incorrect d for SVM classifier " << secIdx << endl;
				return false;
			}
		}
		else
		{
			ds.clear();
			ds.push_back(0.0);
		}

		// Get the penalty parameters to cross-validate
		if (!config.getRealVec(secIdx, "C", Cs))
		{
			cerr << "Error: Missing or incorrect C for SVM classifier " << secIdx << endl;
			return false;
		}
		if (!checkPositive(Cs))
		{
			cerr << "Error: Incorrect C for SVM classifier " << secIdx << endl;
			return false;
		}

		reset(static_cast<unsigned int>(gammas.size() * rs.size() * ds.size() * Cs.size()));

		return true;
	}

	// Print the header information
	void SVMTest::_printHeader(ofstream &file)
	{
		file << "SVM Classifier" << endl
			 << "Strictly supervised: " << (svm.isStrictlySupervised() ? "yes" : "no") << endl
			 << "Compensated projection: " << (svm.usingCompensatedProjection() ? "yes" : "no") << endl
			 << "Kernel: " << svm.getKernelName() << endl;
	}

	// Test the SVM classifier on all available parameters
	bool SVMTest::test(const SimData &data, ofstream &logFile)
	{
		_printHeader(logFile);
		logFile << "d\tr\tgamma\tC\tError rate" << endl;

		unsigned int i = 0; // parameter set count

		vector<int> results(data.getTestSetSize());

		for (vector<double>::const_iterator itr_d = ds.begin(); itr_d != ds.end(); ++itr_d)
		{
			svm.setPara("d", *itr_d);

			for (vector<double>::const_iterator itr_r = rs.begin(); itr_r != rs.end(); ++itr_r)
			{
				svm.setPara("r", *itr_r);

				for (vector<double>::const_iterator itr_gamma = gammas.begin(); itr_gamma != gammas.end(); ++itr_gamma)
				{
					svm.setPara("gamma", *itr_gamma);

					for (vector<double>::const_iterator itrC = Cs.begin(); itrC != Cs.end(); ++itrC)
					{
						svm.setPara("C", *itrC);
						if (!svm.classify(data, results)) return false;

						double errorRate = data.getClassificationError(results);

						// Log the error rate
						logFile << *itr_d << '\t' << *itr_r << '\t' << *itr_gamma << '\t' << *itrC << '\t'
								<< (errorRate * 100.0) << '%' << endl;

						errorSum(i) += errorRate;
						errorSumOfSquares(i++) += (errorRate * errorRate);
					}
				}
			}
		}

		logFile << endl;

		++testCount;

		return true;
	}

	// Clear the results with the best parameters
	void SVMTest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void SVMTest::selectBestPara()
	{
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet

		if (0 == testCount)
		{
			// No cross-validation performed
			latestResult.C = Cs.front();
			latestResult.gamma = gammas.front();
			latestResult.r = rs.front();
			latestResult.d = ds.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);

			// Get the d corresponding to the minimum cross-validation error
			latestResult.d = ds[i / (rs.size() * gammas.size() * Cs.size())];

			// Get the r corresponding to the minimum cross-validation error
			latestResult.r = rs[(i / (gammas.size() * Cs.size())) % rs.size()];

			// Get the gamma corresponding to the minimum cross-validation error
			latestResult.gamma = gammas[(i / Cs.size()) % gammas.size()];

			// Get the C corresponding to the minimum cross-validation error
			latestResult.C = Cs[i % Cs.size()];
		}

		resultsBestPara.push_back(latestResult);
	}

	// Test the SVM classifier with the best parameters
	bool SVMTest::testBestPara(const SimData &data, ofstream &logFile)
	{
		_printHeader(logFile);

		double d = (resultsBestPara.back()).d;
		svm.setPara("d", d);

		double r = (resultsBestPara.back()).r;
		svm.setPara("r", r);

		double gamma = (resultsBestPara.back()).gamma;
		svm.setPara("gamma", gamma);

		double C = (resultsBestPara.back()).C;
		svm.setPara("C", C);

		vector<int> results(data.getTestSetSize());
		if (!svm.classify(data, results)) return false;

		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;

		// Log the error rate
		logFile << "d\tr\tgamma\tC\tError rate" << endl
				<< d << '\t' << r << '\t' << gamma << '\t' << C << '\t'
				<< (errorRate * 100.0) << '%' << endl << endl;

		return true;
	}

	// Report the test results of all available parameters
	void SVMTest::reportResults(ofstream &file, bool verbose)
	{
		_printHeader(file);

		if (verbose)
		{
			// Report the results of all the parameter sets
			file << "d\tr\tgamma\tC\tError rate\tStandard deviation" << endl;

			unsigned int i = 0;
			for (vector<double>::const_iterator itr_d = ds.begin(); itr_d != ds.end(); ++itr_d)
			{
				for (vector<double>::const_iterator itr_r = rs.begin(); itr_r != rs.end(); ++itr_r)
				{
					for (vector<double>::const_iterator itr_gamma = gammas.begin();
						 itr_gamma != gammas.end(); ++itr_gamma)
					{
						for (vector<double>::const_iterator itrC = Cs.begin(); itrC != Cs.end(); ++itrC)
						{
							file << *itr_d << '\t' << *itr_r << '\t' << *itr_gamma << '\t' << *itrC << '\t'
								 << (getError(i) * 100.0) << "%\t" << (getErrorStdDev(i) * 100.0) << '%' << endl;
							++i;
						}
					}
				}
			}
		}
		else
		{
			// Report the best result only
			file << "gamma: " << (resultsBestPara.back()).gamma << endl
				 << "r: " << (resultsBestPara.back()).r << endl
				 << "d: " << (resultsBestPara.back()).d << endl
				 << "C: " << (resultsBestPara.back()).C << endl
				 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << '%' << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << '%' << endl;
		}

		file << endl;
	}

	// Report the test results with the best parameters
	void SVMTest::reportBestParaResults(ofstream &file, bool verbose)
	{
		_printHeader(file);

		file << "gamma: " << (resultsBestPara.back()).gamma << endl
			 << "r: " << (resultsBestPara.back()).r << endl
			 << "d: " << (resultsBestPara.back()).d << endl
			 << "C: " << (resultsBestPara.back()).C << endl;

		if (verbose)
		{
			file << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << '%' << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << '%' << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << '%' << endl;
		}
		else
		{
			file << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << '%' << endl;
		}

		file << endl;
	}

	// Summarize the results of all the tests with the best parameters
	void SVMTest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		_printHeader(file);

		if (verbose)
		{
			file << "d\tr\tgamma\tC\tError rate" << endl;
		}

		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		RealMatrix dVec(1, n);
		RealMatrix rVec(1, n);
		RealMatrix gammaVec(1, n);
		RealMatrix CVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			dVec(i) = itr->d;
			rVec(i) = itr->r;
			gammaVec(i) = itr->gamma;
			CVec(i) = itr->C;
			errorVec(i) = itr->testError;

			if (verbose)
			{
				file << itr->d << '\t' << itr->r << '\t' << itr->gamma << '\t'
					 << itr->C << '\t' << (itr->testError * 100.0) << '%' << endl;
			}
		}

		file << "Average gamma: " << gammaVec.mean() << " (" << gammaVec.std() << ")" << endl
			 << "Average r: " << rVec.mean() << " (" << rVec.std() << ")" << endl
			 << "Average d: " << dVec.mean() << " (" << dVec.std() << ")" << endl
			 << "Average C: " << CVec.mean() << " (" << CVec.std() << ")" << endl
			 << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
