// File: SimEmbed.cpp
// Purpose: Implementation of the SimEmbed class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 4/11/2008 6:06:25 PM

#include <cmath>
#include <iostream>
#include "SimEmbed.h"

using namespace std;

namespace idl
{
	// Initialize the cutoff threshold for determining the embedding dimension
	const double SimEmbed::TH = 1e-6;

	// Embed the similarity data into a Euclidean space
	bool SimEmbed::embed(const RealMatrix &simMat, RealMatrix &embeddedSamples)
	{
		// The similarity matrix to embed must be a square matrix
		if (!simMat.isSquare())
		{
			cerr << "Error: Rectangular similarity matrix to embed" << endl;
			return false;
		}

		unsigned int m = simMat.getRowNum();

		RealMatrix A = simMat + simMat.transpose();
		A *= 0.5; // A = (S + S') / 2

		RealMatrix eigVecs;				// V'
		RealMatrix eigVals;				// D
		A.eig(eigVecs, eigVals);		// Eigendecomposition: A = VDV'

		RealMatrix d = eigVals;			// To store the modified spectrum
		vector<bool> signs(m, false);	// To store the signs of the eigenvalues (false: "+")

		double minEigVal = eigVals.min();

		if (minEigVal < 0.0)
		{
			if (SPECTRUM_SHIFT == spectrumTransformMode)
			{
				d -= minEigVal; // Shift the eigenvalues by a constant
			}
			else if (SPECTRUM_CLIP == spectrumTransformMode)
			{
				for (unsigned int i = 0; i < m; ++i)
				{
					// Clip the negative eigenvalues to zero
					if (d(i) < 0.0) d(i) = 0.0;
				}
			}
			else // SPECTRUM_FLIP == spectrumTransformMode
			{
				for (unsigned int i = 0; i < m; ++i)
				{
					// Flip the sign of the negative eigenvalues
					if (d(i) < 0.0)
					{
						d(i) = -d(i);
						signs[i] = true; // true: "-"
					}
				}
			}
		}

		vector<unsigned int> dims;

		for (unsigned int i = 0; i < m; ++i)
		{
			if (d(i) > TH)
			{
				dims.push_back(i);
				d(i) = sqrt(d(i));	// D = D^(1/2)
			}
		}

		// Dimension of the embedding space
		unsigned int n = static_cast<unsigned int>(dims.size());

		if (0 == n)
		{
			cerr << "Error: All zero similarity matrix to embed" << endl;
			return false;
		}

		embeddedSamples.resize(m, n);	// X' = VD^(1/2)
		projMat.resize(m, n);			// P  = VD^(-1/2)

		for (unsigned int i = 0; i < m; ++i)
		{
			for (unsigned int j = 0; j < n; ++j)
			{
				embeddedSamples(i, j) = eigVecs(dims[j], i) * d(dims[j]);

				double alpha = d(dims[j]);
				if (compensationFlag)
				{
					// Compensated projection
					if (SPECTRUM_SHIFT == spectrumTransformMode)
					{
						if (abs(eigVals(dims[j])) > TH)
						{
							alpha = eigVals(dims[j]) / d(dims[j]);
						}
					}
					else if (SPECTRUM_FLIP == spectrumTransformMode)
					{
						if (signs[dims[j]]) alpha = -alpha;
					}
				}
				projMat(i, j) = eigVecs(dims[j], i) / alpha;
			}
		}

		return true;
	}

	// Project the similarity data into the embedding space
	bool SimEmbed::project(const RealMatrix &simMat, RealMatrix &projectedSamples) const
	{
		if (simMat.getColumnNum() != projMat.getRowNum())
		{
			cerr << "Error: Dimension mismatch between similarity matrix and projection matrix"
				 << endl;
			return false;
		}

		projectedSamples = simMat * projMat; // X' = SP

		return true;
	}
}
