// File: UIntMatrix.cpp
// Purpose: Implementation of the UIntMatrix class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/24/2008 7:37:44 PM

#include <cmath>
#include <algorithm>
#include <map>
#include "UIntMatrix.h"

using namespace std;

namespace idl
{
	// Get the sum of all the elements
	unsigned int UIntMatrix::sum() const
	{
		unsigned int s = 0;

		for (vector<unsigned int>::const_iterator itr = elements.begin();
			 itr != elements.end(); ++itr)
		{
			s += *itr;
		}

		return s;
	}

	// Get the minimum of all the elements
	unsigned int UIntMatrix::min() const
	{
		vector<unsigned int>::const_iterator itr =
			min_element(elements.begin(), elements.end());

		return *itr;
	}

	// Get the sum of a row
	unsigned int UIntMatrix::sumRow(unsigned int i) const
	{
		unsigned int s = 0;

		unsigned int a = i * nColumns;
		unsigned int b = a + nColumns; // (i + 1) * nColumns

		for (unsigned int k = a; k < b; ++k)
		{
			s += elements[k];
		}

		return s;
	}


	// Get the maximum of a row
	unsigned int UIntMatrix::maxRow(unsigned int i) const
	{
		unsigned int a = i * nColumns;
		unsigned int b = a + nColumns; // (i + 1) * nColumns

		unsigned int v = elements[a];

		for (unsigned int k = a + 1; k < b; ++k)
		{
			if (elements[k] > v)
			{
				v = elements[k];
			}
		}

		return v;
	}

	// Get the minimum of a row
	unsigned int UIntMatrix::minRow(unsigned int i) const
	{
		unsigned int a = i * nColumns;
		unsigned int b = a + nColumns; // (i + 1) * nColumns

		unsigned int v = elements[a];

		for (unsigned int k = a + 1; k < b; ++k)
		{
			if (elements[k] < v)
			{
				v = elements[k];
			}
		}

		return v;
	}

	// Get the sum of a column
	unsigned int UIntMatrix::sumColumn(unsigned int j) const
	{
		unsigned int s = 0;

		unsigned int b = nRows * nColumns;

		for (unsigned int k = j; k < b; k += nColumns)
		{
			s += elements[k];
		}

		return s;
	}

	// Get the maximum of a column
	unsigned int UIntMatrix::maxColumn(unsigned int j) const
	{
		unsigned int b = nRows * nColumns;

		unsigned int v = elements[j];

		for (unsigned int k = j + nColumns; k < b; k += nColumns)
		{
			if (elements[k] > v)
			{
				v = elements[k];
			}
		}

		return v;
	}

	// Get the minimum of a column
	unsigned int UIntMatrix::minColumn(unsigned int j) const
	{
		unsigned int b = nRows * nColumns;

		unsigned int v = elements[j];

		for (unsigned int k = j + nColumns; k < b; k += nColumns)
		{
			if (elements[k] < v)
			{
				v = elements[k];
			}
		}

		return v;
	}

	// Get all unique values
	// Returns a Mx1 UIntMatrix of unique values
	void UIntMatrix::unique(UIntMatrix &uniqueVals) const//vector<unsigned int> *uniqueVals)
	{

		map<unsigned int,int> seen;
		for (vector<unsigned int>::const_iterator itr = elements.begin();
			itr != elements.end(); ++itr)
			seen[(*itr)] = 1;
		vector<unsigned int> uniqueValsVect(seen.size());
		
		unsigned int i = 0;
		for (map<unsigned int,int>::const_iterator itr = seen.begin();
			itr != seen.end(); ++itr)
		{
			uniqueValsVect[i] = itr->first;
			++i;
		}
		
		uniqueVals.resize((unsigned int)uniqueValsVect.size(),1);
		i = 0;
		for(vector<unsigned int>::const_iterator itr = uniqueValsVect.begin();
			itr != uniqueValsVect.end(); ++itr)
		{
			uniqueVals(i++,0) = *itr;
		}
	}

	// Compute the function on each element
	void UIntMatrix::fxn(unsigned int (*func)(unsigned int)) 
	{
		for (vector<unsigned int>::iterator itr = elements.begin();
			 itr != elements.end(); ++itr)
		{
			*itr = (*func)(*itr);
		}
		//return *this;
	}

	// Create a matrix of all ones
	UIntMatrix & UIntMatrix::ones()
	{
		fill(1);

		return *this;
	}

	// Create a matrix of all ones
	UIntMatrix & UIntMatrix::ones(unsigned int m, unsigned int n)
	{
		nRows = m;
		nColumns = n;

		elements.clear();
		elements.resize(m * n, 1);

		return *this;
	}

	// Create a matrix of all ones
	UIntMatrix & UIntMatrix::ones(unsigned int m)
	{
		return ones(m, m);
	}

	// Create a matrix of all zeros
	UIntMatrix & UIntMatrix::zeros()
	{
		fill(0);

		return *this;
	}

	// Create a matrix of all zeros
	UIntMatrix & UIntMatrix::zeros(unsigned int m, unsigned int n)
	{
		nRows = m;
		nColumns = n;

		elements.clear();
		elements.resize(m * n, 0);

		return *this;
	}

	// Create a matrix of all zeros
	UIntMatrix & UIntMatrix::zeros(unsigned int m)
	{
		return zeros(m, m);
	}

	// Create an identity matrix
	UIntMatrix & UIntMatrix::eye()
	{
		zeros();

		unsigned int p = (nRows < nColumns) ? nRows : nColumns;

		for (unsigned int i = 0; i < p; ++i)
		{
			elements[i * (nColumns + 1)] = 1;
		}

		return *this;
	}

	// Create an identity matrix
	UIntMatrix & UIntMatrix::eye(unsigned int m)
	{
		zeros(m);

		for (unsigned int i = 0; i < m; ++i)
		{
			elements[i * (m + 1)] = 1;
		}

		return *this;
	}

	// Create an identity matrix
	UIntMatrix & UIntMatrix::eye(unsigned int m, unsigned int n)
	{
		zeros(m, n);

		unsigned int p = (m < n) ? m : n;

		for (unsigned int i = 0; i < p; ++i)
		{
			elements[i * (n + 1)] = 1;
		}

		return *this;
	}


	UIntMatrix & UIntMatrix::operator += (unsigned int v)
	{
		for (vector<unsigned int>::iterator itr = elements.begin();
			 itr != elements.end(); ++itr)
		{
			*itr += v;
		}

		return *this;
	}

	UIntMatrix & UIntMatrix::operator *= (unsigned int v)
	{
		for (vector<unsigned int>::iterator itr = elements.begin();
			 itr != elements.end(); ++itr)
		{
			*itr *= v;
		}

		return *this;
	}

	UIntMatrix & UIntMatrix::operator += (const UIntMatrix &A)
	{
		if (nRows == A.nRows && nColumns == A.nColumns)
		{
			vector<unsigned int>::iterator itr;
			vector<unsigned int>::const_iterator itrA;
			for (itr = elements.begin(), itrA = A.elements.begin();
				 itr != elements.end() && itrA != A.elements.end();
				 ++itr, ++itrA)
			{
				*itr += *itrA;
			}
		}

		return *this;
	}


	UIntMatrix & UIntMatrix::operator *= (const UIntMatrix &A)
	{
		UIntMatrix B = (*this) * A;

		nRows = B.nRows;
		nColumns = B.nColumns;
		elements = B.elements;

		return *this;
	}

	// Get the trace
	unsigned int UIntMatrix::trace() const
	{
		if (!isSquare())
		{
			return 0;
		}

		unsigned int tr = 0;

		for (unsigned int i = 0; i < nRows; ++i)
		{
			tr += elements[i * nColumns + i];
		}

		return tr;
	}

	UIntMatrix operator + (const UIntMatrix &A, unsigned int v)
	{
		UIntMatrix B(A);
		B += v;

		return B;
	}

	UIntMatrix operator * (const UIntMatrix &A, unsigned int v)
	{
		UIntMatrix B(A);
		B *= v;

		return B;
	}

	UIntMatrix operator + (unsigned int v, const UIntMatrix &A)
	{
		UIntMatrix B(A);
		B += v;

		return B;
	}

	UIntMatrix operator * (unsigned int v, const UIntMatrix &A)
	{
		UIntMatrix B(A);
		B *= v;

		return B;
	}

	UIntMatrix operator + (const UIntMatrix &A, const UIntMatrix &B)
	{
		UIntMatrix C(A);
		C += B;

		return C;
	}

	UIntMatrix operator * (const UIntMatrix &A, const UIntMatrix &B)
	{
		UIntMatrix C;

		if (A.getColumnNum() == B.getRowNum())
		{
			unsigned int m = A.getRowNum();
			unsigned int n = B.getColumnNum();
			unsigned int p = A.getColumnNum();

			C.zeros(m, n);
			
			for (unsigned int i = 0; i < m; ++i)
			{
				for (unsigned int j = 0; j < n; ++j)
				{
					for (unsigned int k = 0; k < p; ++k)
					{
						C(i, j) += A(i, k) * B(k, j);
					}
				}
			}
		}

		return C;
	}

	UIntMatrix UIntMatrix::dot(const UIntMatrix &B)
	{
		UIntMatrix C;

		UIntMatrix A = *this;
		if (A.getColumnNum() == B.getColumnNum() && A.getRowNum() == B.getRowNum())
		{
			unsigned int m = A.getRowNum();
			unsigned int n = B.getColumnNum();

			C.zeros(m, n);
			
			for (unsigned int i = 0; i < m; ++i)
			{
				for (unsigned int j = 0; j < n; ++j)
				{
					C(i, j) += A(i, j) * B(i, j);
					
				}
			}
		}

		return C;
	}

	CvMat *UIntMatrix::dumpCvMat() const
	{
		CvMat *mat = cvCreateMat(nRows, nColumns, CV_64FC1);

		vector<unsigned int>::const_iterator itr = elements.begin();

		for (unsigned int i = 0; i < nRows; ++i)
		{
			for (unsigned int j = 0; j < nColumns; ++j)
			{
				cvmSet(mat, i, j, *itr++);
			}
		}

		return mat;
	}

}
