// File: UIntMatrix.h
// Purpose: Definition of the UIntMatrix class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/24/2008 5:54:15 PM

#ifndef IDL_UIntMatrix_H
#define IDL_UIntMatrix_H

#include "cxcore.h" // OpenCV CXCORE header file
#include "Matrix.h"

namespace idl
{
	/**
	 * Class for unsigned int-valued matrix.
	 * @remarks	This class is derived from the Matrix class template.
	 * @see		Matrix
	 */
	class UIntMatrix : public Matrix<unsigned int>
	{
	public:
		UIntMatrix() : Matrix<unsigned int>() {} /**< Default constructor. */

		/**
		 * Constructor.
		 * @param[in]	m	row number
		 * @param[in]	n	column number
		 */
		UIntMatrix(unsigned int m, unsigned int n)
			: Matrix<unsigned int>(m, n)
		{
		}

		/**
		 * Constructor.
		 * @param[in]	m		row number
		 * @param[in]	n		column number
		 * @param[in]	initVal	initial value for all the elements
		 */
		UIntMatrix(unsigned int m, unsigned int n, unsigned int initVal)
			: Matrix<unsigned int>(m, n, initVal)
		{
		}

		/**
		 * Copy constructor.
		 * @param[in]	mat	unsigned int-valued matrix to copy from
		 */
		UIntMatrix(const UIntMatrix &mat) : Matrix<unsigned int>(mat)
		{
		}

		/**
		 * Conversion constructor.
		 * @param[in]	mat unsigned int-valued matrix of the parent class
		 */
		UIntMatrix(const Matrix<unsigned int> &mat) : Matrix<unsigned int>(mat)
		{
		}

		~UIntMatrix() {} /**< Destructor. */

		/**
		 * Assignment operator.
		 * Copy the contents of another unsigned int-valued matrix.
		 * @param[in]	mat unsigned int-valued matrix to copy from
		 * @return		reference to itself
		 */
		UIntMatrix & operator = (const UIntMatrix &mat)
		{
			nRows = mat.nRows;
			nColumns = mat.nColumns;
			elements = mat.elements;

			return *this;
		}

		unsigned int sum() const;		/**< Get the sum of all the elements. */
		unsigned int max() const;		/**< Get the maximum of all the elements. */
		unsigned int min() const;		/**< Get the minimum of all the elements. */

		unsigned int sumRow(unsigned int i) const;	/**< Get the sum of a row. */
		unsigned int maxRow(unsigned int i) const;	/**< Get the maximum of a row. */
		unsigned int minRow(unsigned int i) const;	/**< Get the minimum of a row. */
		unsigned int meanRow(unsigned int i) const;	/**< Get the mean of a row. */

		unsigned int sumColumn(unsigned int j) const;		/**< Get the sum of a column. */
		unsigned int maxColumn(unsigned int j) const;		/**< Get the maximum of a column. */
		unsigned int minColumn(unsigned int j) const;		/**< Get the minimum of a column. */


		void unique(UIntMatrix &uniqueVals) const;

		void fxn(unsigned int (*func)(unsigned int));

		UIntMatrix dot(const UIntMatrix &B);

		/**
		 * Get the trace.
		 * @return		trace
		 * @attention	Trace is only defined for square matrix.
		 */
		unsigned int trace() const;
		
		//@{
		/** Create an identity matrix. */
		UIntMatrix & eye();
		UIntMatrix & eye(unsigned int m);
		UIntMatrix & eye(unsigned int m, unsigned int n);
		//@}

		//@{
		/** Create a matrix of all ones. */
		UIntMatrix & ones();
		UIntMatrix & ones(unsigned int m);
		UIntMatrix & ones(unsigned int m, unsigned int n);
		//@}

		//@{
		/** Create a matrix of all zeros. */
		UIntMatrix & zeros();
		UIntMatrix & zeros(unsigned int m);
		UIntMatrix & zeros(unsigned int m, unsigned int n);
		//@}

		UIntMatrix & negate(); /**< Negate all the elements. */
		UIntMatrix & operator += (unsigned int v);
		UIntMatrix & operator *= (unsigned int v);
		UIntMatrix & operator += (const UIntMatrix &A);
		UIntMatrix & operator *= (const UIntMatrix &A);

		CvMat *dumpCvMat() const;
	};

	UIntMatrix operator + (const UIntMatrix &A, unsigned int v);
	UIntMatrix operator * (const UIntMatrix &A, unsigned int v);


	UIntMatrix operator + (unsigned int v, const UIntMatrix &A);
	UIntMatrix operator * (unsigned int v, const UIntMatrix &A);

	UIntMatrix operator + (const UIntMatrix &A, const UIntMatrix &B);

	UIntMatrix operator * (const UIntMatrix &A, const UIntMatrix &B);
}

#endif // IDL_UIntMatrix_H
