// File: WeightedKNNBN.cpp
// Purpose: Implementation of the WeightedKNNBN class
// Remarks: This class is designed for weighted k-NN classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 6/2/2008 12:04:05 PM

//#include <cstdlib>
//#include <ctime>
#include <cfloat>
#include <map>
#include <vector>
#include <iostream>
#include "CmRandomNumberGenerator.h" // Mersenne Twister (singleton) header file
#include "RealMatrix.h"
#include "WeightedKNNBN.h"

using namespace std;

namespace idl
{
	// Initialize the string representations of the weighting schemes
	const char *WeightedKNNBN::weightNames[] = {
		"uniform", "similarity", "Markov (sim)", "Markov (dissim)",
		"sim-LIME (shift)", "sim-LIME (clip)", "sim-LIME (flip)",
		"sim-LIMV (shift)", "sim-LIMV (clip)", "sim-LIMV (flip)",
		"linear regression (pinv)", "linear regression (shift)",
		"linear regression (clip)", "linear regression (flip)" };

	// Set the weight mode
	bool WeightedKNNBN::setWeightMode(string &strWeightMode)
	{
		if ("uniform" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_UNIFORM;
		}
		else if ("sim" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_SIM;
		}
		else if ("markov_sim" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_MARKOV_SIM;
		}
		else if ("markov_dissim" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_MARKOV_DISSIM;
		}
		else if ("lime_shift" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIME_SHIFT;
		}
		else if ("lime_clip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIME_CLIP;
		}
		else if ("lime_flip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIME_FLIP;
		}
		else if ("limv_shift" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIMV_SHIFT;
		}
		else if ("limv_clip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIMV_CLIP;
		}
		else if ("limv_flip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIMV_FLIP;
		}
		else if ("lr_pinv" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LR_PINV;
		}
		else if ("lr_shift" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LR_SHIFT;
		}
		else if ("lr_clip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LR_CLIP;
		}
		else if ("lr_flip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LR_FLIP;
		}
		else
		{
			return false;
		}

		return true;
	}

	// Classify the test samples of the similarity data set
	bool WeightedKNNBN::classify(const SimData &data, vector<int> &results)
	{
		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size
		sort(parameters.neighborhoodSizes.begin(), parameters.neighborhoodSizes.end());
		if (data.getTrainingSetSize() < parameters.neighborhoodSizes.back())
		{
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}

		// Get the number of test samples
		unsigned int numTestSamples = data.getTestSetSize();

		results.resize(numTestSamples);

		vector<int> uniqueLabels; 
		data.getUniqueLabels(uniqueLabels);
		RealMatrix probabilities;

		WeightedKNN::setWeightMode(parameters.weightMode);
		WeightedKNN::setRegPara(parameters.lambda);
		vector<int> dummyResults;
		
		RealMatrix probMatrix(numTestSamples,(unsigned int)uniqueLabels.size());
		
		for (vector<unsigned int>::const_iterator itrN = parameters.neighborhoodSizes.begin();
				itrN != parameters.neighborhoodSizes.end(); ++itrN)
		{
			WeightedKNN::setNeighborhoodSize(*itrN);
			RealMatrix tmpP;
			WeightedKNN::classify(data,dummyResults,tmpP);
			probMatrix += tmpP;
		
		}

		//probMatrix /= (unsigned int)parameters.neighborhoodSizes.size();
		
		for (unsigned int i = 0; i < numTestSamples; ++i)
		{
			results[i] = maxLabel(uniqueLabels, probMatrix.row(i));
		}
		
		//deprecated
		//for (unsigned int i = 0; i < numTestSamples; ++i)
		//{
		//	//std::cout << "TestSample: " << i << std::endl;
		//	probabilities.zeros(1,(unsigned int)uniqueLabels.size());

		//	for (vector<unsigned int>::const_iterator itrN = parameters.neighborhoodSizes.begin();
		//		itrN != parameters.neighborhoodSizes.end(); ++itrN)
		//	{
		//		WeightedKNN::setNeighborhoodSize(*itrN);
		//		RealMatrix tmpP;
		//		WeightedKNN::getProbabilities(data,dummyResults,tmpP);
		//		probabilities += tmpP;
		//	}
		//	results[i] = maxLabel(uniqueLabels, probabilities);
		//}

		return true;
	}

	bool WeightedKNNBN::classify(const SimData &data, vector<int> &results, RealMatrix &probMatrix)
	{
		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size
		sort(parameters.neighborhoodSizes.begin(), parameters.neighborhoodSizes.end());
		if (data.getTrainingSetSize() < parameters.neighborhoodSizes.back())
		{
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}

		// Get the number of test samples
		unsigned int numTestSamples = data.getTestSetSize();

		results.resize(numTestSamples);

		vector<int> uniqueLabels; 
		data.getUniqueLabels(uniqueLabels);
		RealMatrix probabilities;

		WeightedKNN::setWeightMode(parameters.weightMode);
		WeightedKNN::setRegPara(parameters.lambda);
		vector<int> dummyResults;
		
		probMatrix.resize(numTestSamples,(unsigned int)uniqueLabels.size());
		
		for (vector<unsigned int>::const_iterator itrN = parameters.neighborhoodSizes.begin();
				itrN != parameters.neighborhoodSizes.end(); ++itrN)
		{
			WeightedKNN::setNeighborhoodSize(*itrN);
			RealMatrix tmpP;
			WeightedKNN::classify(data,dummyResults,tmpP);
			probMatrix += tmpP;
		
		}

		probMatrix /= (unsigned int)parameters.neighborhoodSizes.size();
		
		for (unsigned int i = 0; i < numTestSamples; ++i)
		{
			results[i] = maxLabel(uniqueLabels, probMatrix.row(i));
		}
		
		//deprecated
		//for (unsigned int i = 0; i < numTestSamples; ++i)
		//{
		//	//std::cout << "TestSample: " << i << std::endl;
		//	probabilities.zeros(1,(unsigned int)uniqueLabels.size());

		//	for (vector<unsigned int>::const_iterator itrN = parameters.neighborhoodSizes.begin();
		//		itrN != parameters.neighborhoodSizes.end(); ++itrN)
		//	{
		//		WeightedKNN::setNeighborhoodSize(*itrN);
		//		RealMatrix tmpP;
		//		WeightedKNN::getProbabilities(data,dummyResults,tmpP);
		//		probabilities += tmpP;
		//	}
		//	results[i] = maxLabel(uniqueLabels, probabilities);
		//}

		return true;
	}

	// Return the label with the largest probability
	int WeightedKNNBN::maxLabel(const vector<int> &labels, const RealMatrix &probabilities)
	{
		//std::cout << "in max label"; //<<std::cout << probabilities.size()[1];
		double tmpMaxP = -DBL_MAX;
		int tmpMaxLabel = -1;
		unsigned int tmpIdx = 0;
		for (unsigned int tmpIdx = 0; tmpIdx < probabilities.size()[1]; ++tmpIdx)
		{
			if (probabilities(0,tmpIdx) > tmpMaxP)
			{
				tmpMaxP = probabilities(0,tmpIdx);
				tmpMaxLabel = tmpIdx;
			}
		}
		//
		/*for (unsigned int j = 0; j < probabilities.size()[1]; ++j)
		{
			std::cout << probabilities(j) << " ";
		}
		std::cout <<std::endl;
		std::cout <<"label: " << tmpMaxLabel << " " << labels[tmpMaxLabel] <<std::endl;*/
		return labels[tmpMaxLabel];
	}	
}
